from typing import Tuple
import pandas as pd
from pandas import read_csv
import os.path as pth

from aeromaps.models.base import AeromapsModel
from aeromaps.resources import climate_data


class NOxEmissionIndex(AeromapsModel):
    def __init__(self, name="nox_emission_index", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        emission_index_nox_biofuel_2019: float = 0.0,
        emission_index_nox_electrofuel_2019: float = 0.0,
        emission_index_nox_kerosene_2019: float = 0.0,
        emission_index_nox_hydrogen_2019: float = 0.0,
        emission_index_nox_dropin_fuel_evolution: float = 0.0,
        emission_index_nox_hydrogen_evolution: float = 0.0,
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        """NOx emission index calculation using simple method."""

        # Initialization
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "emission_index_nox_biofuel"] = emission_index_nox_biofuel_2019
            self.df.loc[k, "emission_index_nox_electrofuel"] = emission_index_nox_electrofuel_2019
            self.df.loc[k, "emission_index_nox_kerosene"] = emission_index_nox_kerosene_2019
            self.df.loc[k, "emission_index_nox_hydrogen"] = emission_index_nox_hydrogen_2019

        # Calculation
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "emission_index_nox_biofuel"] = self.df.loc[
                k - 1, "emission_index_nox_biofuel"
            ] * (1 + emission_index_nox_dropin_fuel_evolution / 100)
            self.df.loc[k, "emission_index_nox_electrofuel"] = self.df.loc[
                k - 1, "emission_index_nox_electrofuel"
            ] * (1 + emission_index_nox_dropin_fuel_evolution / 100)
            self.df.loc[k, "emission_index_nox_kerosene"] = self.df.loc[
                k - 1, "emission_index_nox_kerosene"
            ] * (1 + emission_index_nox_dropin_fuel_evolution / 100)
            self.df.loc[k, "emission_index_nox_hydrogen"] = self.df.loc[
                k - 1, "emission_index_nox_hydrogen"
            ] * (1 + emission_index_nox_hydrogen_evolution / 100)

        emission_index_nox_biofuel = self.df["emission_index_nox_biofuel"]
        emission_index_nox_electrofuel = self.df["emission_index_nox_electrofuel"]
        emission_index_nox_kerosene = self.df["emission_index_nox_kerosene"]
        emission_index_nox_hydrogen = self.df["emission_index_nox_hydrogen"]

        return (
            emission_index_nox_biofuel,
            emission_index_nox_electrofuel,
            emission_index_nox_kerosene,
            emission_index_nox_hydrogen,
        )


class NOxEmissionIndexComplex(AeromapsModel):
    def __init__(self, name="nox_emission_index_complex", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self.fleet_model = fleet_model

    def compute(
        self,
        emission_index_nox_biofuel_2019: float = 0.0,
        emission_index_nox_electrofuel_2019: float = 0.0,
        emission_index_nox_kerosene_2019: float = 0.0,
        emission_index_nox_hydrogen_2019: float = 0.0,
        ask_long_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        """NOx emission index calculation using fleet renewal models."""

        emission_index_nox_short_range_dropin_fuel = self.fleet_model.df[
            "Short Range:emission_index_nox:dropin_fuel"
        ]
        emission_index_nox_medium_range_dropin_fuel = self.fleet_model.df[
            "Medium Range:emission_index_nox:dropin_fuel"
        ]
        emission_index_nox_long_range_dropin_fuel = self.fleet_model.df[
            "Long Range:emission_index_nox:dropin_fuel"
        ]
        emission_index_nox_short_range_hydrogen = self.fleet_model.df[
            "Short Range:emission_index_nox:hydrogen"
        ]
        emission_index_nox_medium_range_hydrogen = self.fleet_model.df[
            "Medium Range:emission_index_nox:hydrogen"
        ]
        emission_index_nox_long_range_hydrogen = self.fleet_model.df[
            "Long Range:emission_index_nox:hydrogen"
        ]

        # Initialization
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "emission_index_nox_biofuel"] = emission_index_nox_biofuel_2019
            self.df.loc[k, "emission_index_nox_electrofuel"] = emission_index_nox_electrofuel_2019
            self.df.loc[k, "emission_index_nox_kerosene"] = emission_index_nox_kerosene_2019
            self.df.loc[k, "emission_index_nox_hydrogen"] = emission_index_nox_hydrogen_2019

        # Kerosene
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "emission_index_nox_kerosene"] = (
                emission_index_nox_short_range_dropin_fuel.loc[k]
                * ask_short_range_dropin_fuel.loc[k]
                + emission_index_nox_medium_range_dropin_fuel.loc[k]
                * ask_medium_range_dropin_fuel.loc[k]
                + emission_index_nox_long_range_dropin_fuel.loc[k]
                * ask_long_range_dropin_fuel.loc[k]
            ) / (
                ask_short_range_dropin_fuel.loc[k]
                + ask_medium_range_dropin_fuel.loc[k]
                + ask_long_range_dropin_fuel.loc[k]
            )

        # Electrofuel and biofuel
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "emission_index_nox_biofuel"] = (
                emission_index_nox_biofuel_2019
                / emission_index_nox_kerosene_2019
                * self.df.loc[k, "emission_index_nox_kerosene"]
            )
            self.df.loc[k, "emission_index_nox_electrofuel"] = (
                emission_index_nox_electrofuel_2019
                / emission_index_nox_kerosene_2019
                * self.df.loc[k, "emission_index_nox_kerosene"]
            )

        # Hydrogen
        for k in range(self.prospection_start_year, self.end_year + 1):
            if (
                ask_short_range_hydrogen.loc[k]
                + ask_medium_range_hydrogen.loc[k]
                + ask_long_range_hydrogen.loc[k]
                == 0
            ):
                self.df.loc[k, "emission_index_nox_hydrogen"] = self.df.loc[
                    k - 1, "emission_index_nox_hydrogen"
                ]
            else:
                self.df.loc[k, "emission_index_nox_hydrogen"] = (
                    emission_index_nox_short_range_hydrogen.loc[k] * ask_short_range_hydrogen.loc[k]
                    + emission_index_nox_medium_range_hydrogen.loc[k]
                    * ask_medium_range_hydrogen.loc[k]
                    + emission_index_nox_long_range_hydrogen.loc[k] * ask_long_range_hydrogen.loc[k]
                ) / (
                    ask_short_range_hydrogen.loc[k]
                    + ask_medium_range_hydrogen.loc[k]
                    + ask_long_range_hydrogen.loc[k]
                )

        emission_index_nox_biofuel = self.df["emission_index_nox_biofuel"]
        emission_index_nox_electrofuel = self.df["emission_index_nox_electrofuel"]
        emission_index_nox_kerosene = self.df["emission_index_nox_kerosene"]
        emission_index_nox_hydrogen = self.df["emission_index_nox_hydrogen"]

        return (
            emission_index_nox_biofuel,
            emission_index_nox_electrofuel,
            emission_index_nox_kerosene,
            emission_index_nox_hydrogen,
        )


class SootEmissionIndex(AeromapsModel):
    def __init__(self, name="soot_emission_index", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        emission_index_soot_biofuel_2019: float = 0.0,
        emission_index_soot_electrofuel_2019: float = 0.0,
        emission_index_soot_kerosene_2019: float = 0.0,
        emission_index_soot_hydrogen_2019: float = 0.0,
        emission_index_soot_dropin_fuel_evolution: float = 0.0,
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        """Soot emission index calculation using simple method."""

        # Initialization
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "emission_index_soot_biofuel"] = emission_index_soot_biofuel_2019
            self.df.loc[k, "emission_index_soot_electrofuel"] = emission_index_soot_electrofuel_2019
            self.df.loc[k, "emission_index_soot_kerosene"] = emission_index_soot_kerosene_2019
            self.df.loc[k, "emission_index_soot_hydrogen"] = emission_index_soot_hydrogen_2019

        # Calculation
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "emission_index_soot_biofuel"] = self.df.loc[
                k - 1, "emission_index_soot_biofuel"
            ] * (1 + emission_index_soot_dropin_fuel_evolution / 100)
            self.df.loc[k, "emission_index_soot_electrofuel"] = self.df.loc[
                k - 1, "emission_index_soot_electrofuel"
            ] * (1 + emission_index_soot_dropin_fuel_evolution / 100)
            self.df.loc[k, "emission_index_soot_kerosene"] = self.df.loc[
                k - 1, "emission_index_soot_kerosene"
            ] * (1 + emission_index_soot_dropin_fuel_evolution / 100)
            self.df.loc[k, "emission_index_soot_hydrogen"] = emission_index_soot_hydrogen_2019

        emission_index_soot_biofuel = self.df["emission_index_soot_biofuel"]
        emission_index_soot_electrofuel = self.df["emission_index_soot_electrofuel"]
        emission_index_soot_kerosene = self.df["emission_index_soot_kerosene"]
        emission_index_soot_hydrogen = self.df["emission_index_soot_hydrogen"]

        return (
            emission_index_soot_biofuel,
            emission_index_soot_electrofuel,
            emission_index_soot_kerosene,
            emission_index_soot_hydrogen,
        )


class SootEmissionIndexComplex(AeromapsModel):
    def __init__(self, name="soot_emission_index_complex", fleet_model=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self.fleet_model = fleet_model

    def compute(
        self,
        emission_index_soot_biofuel_2019: float = 0.0,
        emission_index_soot_electrofuel_2019: float = 0.0,
        emission_index_soot_kerosene_2019: float = 0.0,
        emission_index_soot_hydrogen_2019: float = 0.0,
        ask_long_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_long_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        ask_medium_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
        ask_short_range_hydrogen: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        """Soot emission index calculation using fleet renewal models."""

        emission_index_soot_short_range_dropin_fuel = self.fleet_model.df[
            "Short Range:emission_index_soot:dropin_fuel"
        ]
        emission_index_soot_medium_range_dropin_fuel = self.fleet_model.df[
            "Medium Range:emission_index_soot:dropin_fuel"
        ]
        emission_index_soot_long_range_dropin_fuel = self.fleet_model.df[
            "Long Range:emission_index_soot:dropin_fuel"
        ]
        emission_index_soot_short_range_hydrogen = self.fleet_model.df[
            "Short Range:emission_index_soot:hydrogen"
        ]
        emission_index_soot_medium_range_hydrogen = self.fleet_model.df[
            "Medium Range:emission_index_soot:hydrogen"
        ]
        emission_index_soot_long_range_hydrogen = self.fleet_model.df[
            "Long Range:emission_index_soot:hydrogen"
        ]

        # Initialization
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "emission_index_soot_biofuel"] = emission_index_soot_biofuel_2019
            self.df.loc[k, "emission_index_soot_electrofuel"] = emission_index_soot_electrofuel_2019
            self.df.loc[k, "emission_index_soot_kerosene"] = emission_index_soot_kerosene_2019
            self.df.loc[k, "emission_index_soot_hydrogen"] = emission_index_soot_hydrogen_2019

        # Kerosene
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "emission_index_soot_kerosene"] = (
                emission_index_soot_short_range_dropin_fuel.loc[k]
                * ask_short_range_dropin_fuel.loc[k]
                + emission_index_soot_medium_range_dropin_fuel.loc[k]
                * ask_medium_range_dropin_fuel.loc[k]
                + emission_index_soot_long_range_dropin_fuel.loc[k]
                * ask_long_range_dropin_fuel.loc[k]
            ) / (
                ask_short_range_dropin_fuel.loc[k]
                + ask_medium_range_dropin_fuel.loc[k]
                + ask_long_range_dropin_fuel.loc[k]
            )

        # Electrofuel and biofuel
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "emission_index_soot_biofuel"] = (
                emission_index_soot_biofuel_2019
                / emission_index_soot_kerosene_2019
                * self.df.loc[k, "emission_index_soot_kerosene"]
            )
            self.df.loc[k, "emission_index_soot_electrofuel"] = (
                emission_index_soot_electrofuel_2019
                / emission_index_soot_kerosene_2019
                * self.df.loc[k, "emission_index_soot_kerosene"]
            )

        # Hydrogen
        for k in range(self.prospection_start_year, self.end_year + 1):
            if (
                ask_short_range_hydrogen.loc[k]
                + ask_medium_range_hydrogen.loc[k]
                + ask_long_range_hydrogen.loc[k]
                == 0
            ):
                self.df.loc[k, "emission_index_soot_hydrogen"] = self.df.loc[
                    k - 1, "emission_index_soot_hydrogen"
                ]
            else:
                self.df.loc[k, "emission_index_soot_hydrogen"] = (
                    emission_index_soot_short_range_hydrogen.loc[k]
                    * ask_short_range_hydrogen.loc[k]
                    + emission_index_soot_medium_range_hydrogen.loc[k]
                    * ask_medium_range_hydrogen.loc[k]
                    + emission_index_soot_long_range_hydrogen.loc[k]
                    * ask_long_range_hydrogen.loc[k]
                ) / (
                    ask_short_range_hydrogen.loc[k]
                    + ask_medium_range_hydrogen.loc[k]
                    + ask_long_range_hydrogen.loc[k]
                )

        emission_index_soot_biofuel = self.df["emission_index_soot_biofuel"]
        emission_index_soot_electrofuel = self.df["emission_index_soot_electrofuel"]
        emission_index_soot_kerosene = self.df["emission_index_soot_kerosene"]
        emission_index_soot_hydrogen = self.df["emission_index_soot_hydrogen"]

        return (
            emission_index_soot_biofuel,
            emission_index_soot_electrofuel,
            emission_index_soot_kerosene,
            emission_index_soot_hydrogen,
        )


class NonCO2Emissions(AeromapsModel):
    def __init__(self, name="non_co2_emissions", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        # Load dataset
        historical_dataset_path = pth.join(
            climate_data.__path__[0], "temperature_historical_dataset.csv"
        )
        historical_dataset_df = read_csv(historical_dataset_path, delimiter=";", header=None)
        self.historical_dataset = historical_dataset_df.values

    def compute(
        self,
        emission_index_nox_biofuel: pd.Series = pd.Series(dtype="float64"),
        emission_index_nox_electrofuel: pd.Series = pd.Series(dtype="float64"),
        emission_index_nox_kerosene: pd.Series = pd.Series(dtype="float64"),
        emission_index_nox_hydrogen: pd.Series = pd.Series(dtype="float64"),
        emission_index_soot_biofuel: pd.Series = pd.Series(dtype="float64"),
        emission_index_soot_electrofuel: pd.Series = pd.Series(dtype="float64"),
        emission_index_soot_kerosene: pd.Series = pd.Series(dtype="float64"),
        emission_index_soot_hydrogen: pd.Series = pd.Series(dtype="float64"),
        emission_index_h2o_biofuel: float = 0.0,
        emission_index_h2o_electrofuel: float = 0.0,
        emission_index_h2o_kerosene: float = 0.0,
        emission_index_h2o_hydrogen: float = 0.0,
        emission_index_sulfur_biofuel: float = 0.0,
        emission_index_sulfur_electrofuel: float = 0.0,
        emission_index_sulfur_kerosene: float = 0.0,
        emission_index_sulfur_hydrogen: float = 0.0,
        lhv_kerosene: float = 0.0,
        lhv_biofuel: float = 0.0,
        lhv_electrofuel: float = 0.0,
        lhv_hydrogen: float = 0.0,
        energy_consumption_kerosene: pd.Series = pd.Series(dtype="float64"),
        energy_consumption_biofuel: pd.Series = pd.Series(dtype="float64"),
        energy_consumption_electrofuel: pd.Series = pd.Series(dtype="float64"),
        energy_consumption_hydrogen: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        """Non-CO2 emissions calculation."""

        ## Initialization
        historical_nox_emissions_for_temperature = self.historical_dataset[:, 2]
        historical_h2o_emissions_for_temperature = self.historical_dataset[:, 3]
        historical_soot_emissions_for_temperature = self.historical_dataset[:, 4]
        historical_sulfur_emissions_for_temperature = self.historical_dataset[:, 5]

        # Calculation
        for k in range(self.climate_historic_start_year, self.historic_start_year):
            self.df_climate.loc[k, "soot_emissions"] = historical_soot_emissions_for_temperature[
                k - self.climate_historic_start_year
            ]
            self.df_climate.loc[k, "h2o_emissions"] = historical_h2o_emissions_for_temperature[
                k - self.climate_historic_start_year
            ]
            self.df_climate.loc[k, "nox_emissions"] = historical_nox_emissions_for_temperature[
                k - self.climate_historic_start_year
            ]
            self.df_climate.loc[
                k, "sulfur_emissions"
            ] = historical_sulfur_emissions_for_temperature[k - self.climate_historic_start_year]

        for k in range(self.historic_start_year, self.end_year + 1):
            self.df_climate.loc[k, "soot_emissions"] = (
                emission_index_soot_biofuel.loc[k]
                / 10**9
                * energy_consumption_biofuel.loc[k]
                / lhv_biofuel
                + emission_index_soot_electrofuel.loc[k]
                / 10**9
                * energy_consumption_electrofuel.loc[k]
                / lhv_electrofuel
                + emission_index_soot_kerosene.loc[k]
                / 10**9
                * energy_consumption_kerosene.loc[k]
                / lhv_kerosene
                + emission_index_soot_hydrogen.loc[k]
                / 10**9
                * energy_consumption_hydrogen.loc[k]
                / lhv_hydrogen
            )
            self.df_climate.loc[k, "h2o_emissions"] = (
                emission_index_h2o_biofuel
                / 10**9
                * energy_consumption_biofuel.loc[k]
                / lhv_biofuel
                + emission_index_h2o_electrofuel
                / 10**9
                * energy_consumption_electrofuel.loc[k]
                / lhv_electrofuel
                + emission_index_h2o_kerosene
                / 10**9
                * energy_consumption_kerosene.loc[k]
                / lhv_kerosene
                + emission_index_h2o_hydrogen
                / 10**9
                * energy_consumption_hydrogen.loc[k]
                / lhv_hydrogen
            )
            self.df_climate.loc[k, "nox_emissions"] = (
                emission_index_nox_biofuel.loc[k]
                / 10**9
                * energy_consumption_biofuel.loc[k]
                / lhv_biofuel
                + emission_index_nox_electrofuel.loc[k]
                / 10**9
                * energy_consumption_electrofuel.loc[k]
                / lhv_electrofuel
                + emission_index_nox_kerosene.loc[k]
                / 10**9
                * energy_consumption_kerosene.loc[k]
                / lhv_kerosene
                + emission_index_nox_hydrogen.loc[k]
                / 10**9
                * energy_consumption_hydrogen.loc[k]
                / lhv_hydrogen
            )
            self.df_climate.loc[k, "sulfur_emissions"] = (
                emission_index_sulfur_biofuel
                / 10**9
                * energy_consumption_biofuel.loc[k]
                / lhv_biofuel
                + emission_index_sulfur_electrofuel
                / 10**9
                * energy_consumption_electrofuel.loc[k]
                / lhv_electrofuel
                + emission_index_sulfur_kerosene
                / 10**9
                * energy_consumption_kerosene.loc[k]
                / lhv_kerosene
                + emission_index_sulfur_hydrogen
                / 10**9
                * energy_consumption_hydrogen.loc[k]
                / lhv_hydrogen
            )

        soot_emissions = self.df_climate.loc[:, "soot_emissions"]
        h2o_emissions = self.df_climate.loc[:, "h2o_emissions"]
        nox_emissions = self.df_climate.loc[:, "nox_emissions"]
        sulfur_emissions = self.df_climate.loc[:, "sulfur_emissions"]

        return soot_emissions, h2o_emissions, nox_emissions, sulfur_emissions



from typing import Tuple

import pandas as pd

from aeromaps.models.base import (
    AeromapsModel,
    AeromapsInterpolationFunction,
    AeromapsLevelingFunction,
)


class LevelCarbonOffset(AeromapsModel):
    def __init__(self, name="level_carbon_offset", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        co2_emissions: pd.Series = pd.Series(dtype="float64"),
        carbon_offset_baseline_level_vs_2019_reference_periods: list = [],
        carbon_offset_baseline_level_vs_2019_reference_periods_values: list = [],
    ) -> Tuple[pd.Series, pd.Series]:

        carbon_offset_baseline_level_vs_2019 = AeromapsLevelingFunction(
            self,
            carbon_offset_baseline_level_vs_2019_reference_periods,
            carbon_offset_baseline_level_vs_2019_reference_periods_values,
            model_name=self.name,
        )
        self.df.loc[
            :, "carbon_offset_baseline_level_vs_2019"
        ] = carbon_offset_baseline_level_vs_2019

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "level_carbon_offset"] = 0.0

        for k in range(self.prospection_start_year, self.end_year + 1):
            if (
                co2_emissions.loc[k]
                > co2_emissions.loc[2019]
                * self.df.loc[k, "carbon_offset_baseline_level_vs_2019"]
                / 100
            ):
                self.df.loc[k, "level_carbon_offset"] = (
                    co2_emissions.loc[k]
                    - co2_emissions.loc[2019]
                    * self.df.loc[k, "carbon_offset_baseline_level_vs_2019"]
                    / 100
                )
            else:
                self.df.loc[k, "level_carbon_offset"] = 0.0

        level_carbon_offset = self.df["level_carbon_offset"]

        return (carbon_offset_baseline_level_vs_2019, level_carbon_offset)


class ResidualCarbonOffset(AeromapsModel):
    def __init__(self, name="residual_carbon_offset", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        co2_emissions: pd.Series = pd.Series(dtype="float64"),
        level_carbon_offset: pd.Series = pd.Series(dtype="float64"),
        residual_carbon_offset_share_reference_years: list = [],
        residual_carbon_offset_share_reference_years_values: list = [],
    ) -> Tuple[pd.Series, pd.Series]:

        residual_carbon_offset_share_prospective = AeromapsInterpolationFunction(
            self,
            residual_carbon_offset_share_reference_years,
            residual_carbon_offset_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "residual_carbon_offset_share"] = residual_carbon_offset_share_prospective
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "residual_carbon_offset_share"] = 0.0
        residual_carbon_offset_share = self.df["residual_carbon_offset_share"]

        for k in range(self.historic_start_year, self.end_year + 1):
            self.df.loc[k, "residual_carbon_offset"] = (
                self.df.loc[k, "residual_carbon_offset_share"]
                / 100
                * (co2_emissions.loc[k] - level_carbon_offset.loc[k])
            )

        residual_carbon_offset = self.df["residual_carbon_offset"]

        return (residual_carbon_offset_share, residual_carbon_offset)


class CarbonOffset(AeromapsModel):
    def __init__(self, name="carbon_offset", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        level_carbon_offset: pd.Series = pd.Series(dtype="float64"),
        residual_carbon_offset: pd.Series = pd.Series(dtype="float64"),
    ) -> pd.Series:

        carbon_offset = level_carbon_offset + residual_carbon_offset

        self.df.loc[:, "carbon_offset"] = carbon_offset

        return carbon_offset


class CumulativeCarbonOffset(AeromapsModel):
    def __init__(self, name="cumulative_carbon_offset", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        carbon_offset: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series]:

        self.df.loc[self.prospection_start_year - 1, "cumulative_carbon_offset"] = 0.0
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "cumulative_carbon_offset"] = (
                self.df.loc[k - 1, "cumulative_carbon_offset"] + carbon_offset.loc[k] / 1000
            )

        cumulative_carbon_offset = self.df["cumulative_carbon_offset"]

        return cumulative_carbon_offset



from typing import Tuple

import numpy as np
import pandas as pd

from aeromaps.models.base import AeromapsModel


class OperationsContrailsSimple(AeromapsModel):
    def __init__(self, name="operations_contrails_simple", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        operations_contrails_final_gain: float,
        operations_contrails_final_overconsumption: float,
        operations_contrails_start_year: float,
        operations_contrails_duration: float,
    ) -> Tuple[pd.Series, pd.Series]:
        """Operations contrails gain for ERF calculation."""

        transition_year = operations_contrails_start_year + operations_contrails_duration / 2
        operations_contrails_limit = 0.02 * operations_contrails_final_gain
        operations_parameter = np.log(100 / 2 - 1) / (operations_contrails_duration / 2)
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "operations_contrails_gain"] = 0
            self.df.loc[k, "operations_contrails_overconsumption"] = 0
        for k in range(self.prospection_start_year, self.end_year + 1):
            if (
                operations_contrails_final_gain
                / (1 + np.exp(-operations_parameter * (k - transition_year)))
                < operations_contrails_limit
            ):
                self.df.loc[k, "operations_contrails_gain"] = 0
                self.df.loc[k, "operations_contrails_overconsumption"] = 0
            else:
                self.df.loc[k, "operations_contrails_gain"] = operations_contrails_final_gain / (
                    1 + np.exp(-operations_parameter * (k - transition_year))
                )
                self.df.loc[
                    k, "operations_contrails_overconsumption"
                ] = operations_contrails_final_overconsumption / (
                    1 + np.exp(-operations_parameter * (k - transition_year))
                )

        operations_contrails_gain = self.df["operations_contrails_gain"]
        operations_contrails_overconsumption = self.df["operations_contrails_overconsumption"]

        return operations_contrails_gain, operations_contrails_overconsumption


class FuelEffectCorrectionContrails(AeromapsModel):
    def __init__(self, name="fuel_effect_correction_contrails", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        total_aircraft_distance_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        total_aircraft_distance: pd.Series = pd.Series(dtype="float64"),
        total_aircraft_distance_hydrogen: pd.Series = pd.Series(dtype="float64"),
        total_aircraft_distance_electric: pd.Series = pd.Series(dtype="float64"),
        biofuel_share: pd.Series = pd.Series(dtype="float64"),
        electrofuel_share: pd.Series = pd.Series(dtype="float64"),
        kerosene_share: pd.Series = pd.Series(dtype="float64"),
        emission_index_number_particles_biofuel: float = 0.0,
        emission_index_number_particles_electrofuel: float = 0.0,
        emission_index_number_particles_kerosene: float = 0.0,
        contrails_relative_effect_hydrogen_wrt_kerosene: float = 0.0,
    ) -> Tuple[pd.Series]:

        """Fuel effect on contrails for ERF calculation."""

        fuel_effect_correction_contrails = (
            total_aircraft_distance_dropin_fuel
            / total_aircraft_distance
            * (
                kerosene_share
                / 100
                * np.sqrt(
                    emission_index_number_particles_kerosene
                    / emission_index_number_particles_kerosene
                )
                + biofuel_share
                / 100
                * np.sqrt(
                    emission_index_number_particles_biofuel
                    / emission_index_number_particles_kerosene
                )
                + electrofuel_share
                / 100
                * np.sqrt(
                    emission_index_number_particles_electrofuel
                    / emission_index_number_particles_kerosene
                )
            )
            + total_aircraft_distance_hydrogen
            / total_aircraft_distance
            * contrails_relative_effect_hydrogen_wrt_kerosene
            + total_aircraft_distance_electric / total_aircraft_distance * 0
        )

        self.df["fuel_effect_correction_contrails"] = fuel_effect_correction_contrails

        return fuel_effect_correction_contrails


class WithoutFuelEffectCorrectionContrails(AeromapsModel):
    def __init__(self, name="without_fuel_effect_correction_contrails", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
    ) -> Tuple[pd.Series]:

        """Fuel effect on contrails for ERF calculation."""

        for k in range(self.historic_start_year, self.end_year + 1):
            self.df.loc[k, "fuel_effect_correction_contrails"] = 1

        fuel_effect_correction_contrails = self.df["fuel_effect_correction_contrails"]

        return fuel_effect_correction_contrails



from typing import Tuple

import pandas as pd

from aeromaps.models.base import AeromapsModel, AeromapsInterpolationFunction


class BiofuelEfficiency(AeromapsModel):
    def __init__(self, name="biofuel_efficiency", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        biofuel_ft_efficiency_reference_years: list = [],
        biofuel_ft_efficiency_reference_years_values: list = [],
        biofuel_atj_efficiency_reference_years: list = [],
        biofuel_atj_efficiency_reference_years_values: list = [],
        biofuel_hefa_oil_efficiency_reference_years: list = [],
        biofuel_hefa_oil_efficiency_reference_years_values: list = [],
        biofuel_hefa_fuel_efficiency_reference_years: list = [],
        biofuel_hefa_fuel_efficiency_reference_years_values: list = [],
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        """Biofuel production efficiency calculation using interpolation functions"""

        # FT
        biofuel_ft_efficiency = AeromapsInterpolationFunction(
            self,
            biofuel_ft_efficiency_reference_years,
            biofuel_ft_efficiency_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_efficiency"] = biofuel_ft_efficiency

        # ATJ
        biofuel_atj_efficiency = AeromapsInterpolationFunction(
            self,
            biofuel_atj_efficiency_reference_years,
            biofuel_atj_efficiency_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_atj_efficiency"] = biofuel_atj_efficiency

        # HEFA OIL
        biofuel_hefa_oil_efficiency = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_oil_efficiency_reference_years,
            biofuel_hefa_oil_efficiency_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_oil_efficiency"] = biofuel_hefa_oil_efficiency

        # HEFA FUEL
        biofuel_hefa_fuel_efficiency = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_fuel_efficiency_reference_years,
            biofuel_hefa_fuel_efficiency_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_fuel_efficiency"] = biofuel_hefa_fuel_efficiency

        return (
            biofuel_ft_efficiency,
            biofuel_atj_efficiency,
            biofuel_hefa_oil_efficiency,
            biofuel_hefa_fuel_efficiency,
        )


class ElectricityBasedFuelEfficiency(AeromapsModel):
    def __init__(self, name="electricity_based_fuel_efficiency", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        electrolysis_efficiency_reference_years: list = [],
        electrolysis_efficiency_reference_years_values: list = [],
        liquefaction_efficiency_reference_years: list = [],
        liquefaction_efficiency_reference_years_values: list = [],
        electrofuel_hydrogen_efficiency_reference_years: list = [],
        electrofuel_hydrogen_efficiency_reference_years_values: list = [],
    ) -> Tuple[pd.Series, pd.Series, pd.Series]:
        """Hydrogen and electrofuel production efficiency calculation using interpolation functions"""

        # Electrolysis
        electrolysis_efficiency = AeromapsInterpolationFunction(
            self,
            electrolysis_efficiency_reference_years,
            electrolysis_efficiency_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "electrolysis_efficiency"] = electrolysis_efficiency

        # Liquefaction
        liquefaction_efficiency = AeromapsInterpolationFunction(
            self,
            liquefaction_efficiency_reference_years,
            liquefaction_efficiency_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "liquefaction_efficiency"] = liquefaction_efficiency

        # Electrofuel from hydrogen
        electrofuel_hydrogen_efficiency = AeromapsInterpolationFunction(
            self,
            electrofuel_hydrogen_efficiency_reference_years,
            electrofuel_hydrogen_efficiency_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "electrofuel_hydrogen_efficiency"] = electrofuel_hydrogen_efficiency

        return electrolysis_efficiency, liquefaction_efficiency, electrofuel_hydrogen_efficiency



from typing import Tuple

import numpy as np
import pandas as pd
import os.path as pth

from aeromaps.models.base import (
    AeromapsModel,
    GWPStarEquivalentEmissionsFunction,
    AbsoluteGlobalWarmingPotentialCO2Function,
    RunFair,
)
from aeromaps.resources.climate_data import RCP


class TemperatureGWPStar(AeromapsModel):
    def __init__(self, name="temperature_gwpstar", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        contrails_gwpstar_variation_duration: float = 0.0,
        contrails_gwpstar_s_coefficient: float = 0.0,
        nox_short_term_o3_increase_gwpstar_variation_duration: float = 0.0,
        nox_short_term_o3_increase_gwpstar_s_coefficient: float = 0.0,
        nox_long_term_o3_decrease_gwpstar_variation_duration: float = 0.0,
        nox_long_term_o3_decrease_gwpstar_s_coefficient: float = 0.0,
        nox_ch4_decrease_gwpstar_variation_duration: float = 0.0,
        nox_ch4_decrease_gwpstar_s_coefficient: float = 0.0,
        nox_stratospheric_water_vapor_decrease_gwpstar_variation_duration: float = 0.0,
        nox_stratospheric_water_vapor_decrease_gwpstar_s_coefficient: float = 0.0,
        soot_gwpstar_variation_duration: float = 0.0,
        soot_gwpstar_s_coefficient: float = 0.0,
        h2o_gwpstar_variation_duration: float = 0.0,
        h2o_gwpstar_s_coefficient: float = 0.0,
        sulfur_gwpstar_variation_duration: float = 0.0,
        sulfur_gwpstar_s_coefficient: float = 0.0,
        contrails_erf: pd.Series = pd.Series(dtype="float64"),
        nox_short_term_o3_increase_erf: pd.Series = pd.Series(dtype="float64"),
        nox_long_term_o3_decrease_erf: pd.Series = pd.Series(dtype="float64"),
        nox_ch4_decrease_erf: pd.Series = pd.Series(dtype="float64"),
        nox_stratospheric_water_vapor_decrease_erf: pd.Series = pd.Series(dtype="float64"),
        soot_erf: pd.Series = pd.Series(dtype="float64"),
        h2o_erf: pd.Series = pd.Series(dtype="float64"),
        sulfur_erf: pd.Series = pd.Series(dtype="float64"),
        co2_erf: pd.Series = pd.Series(dtype="float64"),
        total_erf: pd.Series = pd.Series(dtype="float64"),
        co2_emissions: pd.Series = pd.Series(dtype="float64"),
        tcre_coefficient: float = 0.0,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """Temperature calculation using equivalent emissions (with GWP* method) and TCRE."""

        # EQUIVALENT EMISSIONS

        ## Contrails
        contrails_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=contrails_erf,
            gwpstar_variation_duration=contrails_gwpstar_variation_duration,
            gwpstar_s_coefficient=contrails_gwpstar_s_coefficient,
        )
        self.df_climate["contrails_equivalent_emissions"] = contrails_equivalent_emissions

        ## NOx short-term O3 increase
        nox_short_term_o3_increase_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=nox_short_term_o3_increase_erf,
            gwpstar_variation_duration=nox_short_term_o3_increase_gwpstar_variation_duration,
            gwpstar_s_coefficient=nox_short_term_o3_increase_gwpstar_s_coefficient,
        )
        self.df_climate[
            "nox_short_term_o3_increase_equivalent_emissions"
        ] = nox_short_term_o3_increase_equivalent_emissions

        ## NOx long-term O3 decrease
        nox_long_term_o3_decrease_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=nox_long_term_o3_decrease_erf,
            gwpstar_variation_duration=nox_long_term_o3_decrease_gwpstar_variation_duration,
            gwpstar_s_coefficient=nox_long_term_o3_decrease_gwpstar_s_coefficient,
        )
        self.df_climate[
            "nox_long_term_o3_decrease_equivalent_emissions"
        ] = nox_long_term_o3_decrease_equivalent_emissions

        ## NOx CH4 decrease
        nox_ch4_decrease_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=nox_ch4_decrease_erf,
            gwpstar_variation_duration=nox_ch4_decrease_gwpstar_variation_duration,
            gwpstar_s_coefficient=nox_ch4_decrease_gwpstar_s_coefficient,
        )
        self.df_climate[
            "nox_ch4_decrease_equivalent_emissions"
        ] = nox_ch4_decrease_equivalent_emissions

        ## NOx stratospheric water vapor decrease
        nox_stratospheric_water_vapor_decrease_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=nox_stratospheric_water_vapor_decrease_erf,
            gwpstar_variation_duration=nox_stratospheric_water_vapor_decrease_gwpstar_variation_duration,
            gwpstar_s_coefficient=nox_stratospheric_water_vapor_decrease_gwpstar_s_coefficient,
        )
        self.df_climate[
            "nox_stratospheric_water_vapor_decrease_equivalent_emissions"
        ] = nox_stratospheric_water_vapor_decrease_equivalent_emissions

        ## Soot
        soot_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=soot_erf,
            gwpstar_variation_duration=soot_gwpstar_variation_duration,
            gwpstar_s_coefficient=soot_gwpstar_s_coefficient,
        )
        self.df_climate["soot_equivalent_emissions"] = soot_equivalent_emissions

        ## H2O
        h2o_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=h2o_erf,
            gwpstar_variation_duration=h2o_gwpstar_variation_duration,
            gwpstar_s_coefficient=h2o_gwpstar_s_coefficient,
        )
        self.df_climate["h2o_equivalent_emissions"] = h2o_equivalent_emissions

        ## Sulfur
        sulfur_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=sulfur_erf,
            gwpstar_variation_duration=sulfur_gwpstar_variation_duration,
            gwpstar_s_coefficient=sulfur_gwpstar_s_coefficient,
        )
        self.df_climate["sulfur_equivalent_emissions"] = sulfur_equivalent_emissions

        ## Total
        non_co2_equivalent_emissions = (
            contrails_equivalent_emissions
            + nox_short_term_o3_increase_equivalent_emissions
            + nox_long_term_o3_decrease_equivalent_emissions
            + nox_ch4_decrease_equivalent_emissions
            + nox_stratospheric_water_vapor_decrease_equivalent_emissions
            + soot_equivalent_emissions
            + h2o_equivalent_emissions
            + sulfur_equivalent_emissions
        )
        total_equivalent_emissions = co2_emissions + non_co2_equivalent_emissions
        self.df_climate["non_co2_equivalent_emissions"] = non_co2_equivalent_emissions
        self.df_climate["total_equivalent_emissions"] = total_equivalent_emissions

        ## Cumulative CO2, non-CO2 and total equivalent emissions (Gtwe)

        ### From 1940
        self.df_climate.loc[
            self.climate_historic_start_year, "historical_cumulative_co2_emissions"
        ] = (co2_emissions.loc[self.climate_historic_start_year] / 1000)
        self.df_climate.loc[
            self.climate_historic_start_year, "historical_cumulative_contrails_equivalent_emissions"
        ] = (contrails_equivalent_emissions.loc[self.climate_historic_start_year] / 1000)
        self.df_climate.loc[
            self.climate_historic_start_year,
            "historical_cumulative_nox_short_term_o3_increase_equivalent_emissions",
        ] = (
            nox_short_term_o3_increase_equivalent_emissions.loc[self.climate_historic_start_year]
            / 1000
        )
        self.df_climate.loc[
            self.climate_historic_start_year,
            "historical_cumulative_nox_long_term_o3_decrease_equivalent_emissions",
        ] = (
            nox_long_term_o3_decrease_equivalent_emissions.loc[self.climate_historic_start_year]
            / 1000
        )
        self.df_climate.loc[
            self.climate_historic_start_year,
            "historical_cumulative_nox_ch4_decrease_equivalent_emissions",
        ] = (
            nox_ch4_decrease_equivalent_emissions.loc[self.climate_historic_start_year] / 1000
        )
        self.df_climate.loc[
            self.climate_historic_start_year,
            "historical_cumulative_nox_stratospheric_water_vapor_decrease_equivalent_emissions",
        ] = (
            nox_stratospheric_water_vapor_decrease_equivalent_emissions.loc[
                self.climate_historic_start_year
            ]
            / 1000
        )
        self.df_climate.loc[
            self.climate_historic_start_year, "historical_cumulative_soot_equivalent_emissions"
        ] = (soot_equivalent_emissions.loc[self.climate_historic_start_year] / 1000)
        self.df_climate.loc[
            self.climate_historic_start_year, "historical_cumulative_h2o_equivalent_emissions"
        ] = (h2o_equivalent_emissions.loc[self.climate_historic_start_year] / 1000)
        self.df_climate.loc[
            self.climate_historic_start_year, "historical_cumulative_sulfur_equivalent_emissions"
        ] = (sulfur_equivalent_emissions.loc[self.climate_historic_start_year] / 1000)
        self.df_climate.loc[
            self.climate_historic_start_year, "historical_cumulative_non_co2_equivalent_emissions"
        ] = (non_co2_equivalent_emissions[self.climate_historic_start_year] / 1000)
        for k in range(self.climate_historic_start_year + 1, self.end_year + 1):
            self.df_climate.loc[k, "historical_cumulative_co2_emissions"] = (
                self.df_climate.loc[k - 1, "historical_cumulative_co2_emissions"]
                + co2_emissions.loc[k] / 1000
            )
            self.df_climate.loc[k, "historical_cumulative_contrails_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "historical_cumulative_contrails_equivalent_emissions"]
                + contrails_equivalent_emissions.loc[k] / 1000
            )
            self.df_climate.loc[
                k, "historical_cumulative_nox_short_term_o3_increase_equivalent_emissions"
            ] = (
                self.df_climate.loc[
                    k - 1, "historical_cumulative_nox_short_term_o3_increase_equivalent_emissions"
                ]
                + nox_short_term_o3_increase_equivalent_emissions.loc[k] / 1000
            )
            self.df_climate.loc[
                k, "historical_cumulative_nox_long_term_o3_decrease_equivalent_emissions"
            ] = (
                self.df_climate.loc[
                    k - 1, "historical_cumulative_nox_long_term_o3_decrease_equivalent_emissions"
                ]
                + nox_long_term_o3_decrease_equivalent_emissions.loc[k] / 1000
            )
            self.df_climate.loc[
                k, "historical_cumulative_nox_ch4_decrease_equivalent_emissions"
            ] = (
                self.df_climate.loc[
                    k - 1, "historical_cumulative_nox_ch4_decrease_equivalent_emissions"
                ]
                + nox_ch4_decrease_equivalent_emissions.loc[k] / 1000
            )
            self.df_climate.loc[
                k,
                "historical_cumulative_nox_stratospheric_water_vapor_decrease_equivalent_emissions",
            ] = (
                self.df_climate.loc[
                    k - 1,
                    "historical_cumulative_nox_stratospheric_water_vapor_decrease_equivalent_emissions",
                ]
                + nox_stratospheric_water_vapor_decrease_equivalent_emissions.loc[k] / 1000
            )
            self.df_climate.loc[k, "historical_cumulative_soot_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "historical_cumulative_soot_equivalent_emissions"]
                + soot_equivalent_emissions.loc[k] / 1000
            )
            self.df_climate.loc[k, "historical_cumulative_h2o_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "historical_cumulative_h2o_equivalent_emissions"]
                + h2o_equivalent_emissions.loc[k] / 1000
            )
            self.df_climate.loc[k, "historical_cumulative_sulfur_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "historical_cumulative_sulfur_equivalent_emissions"]
                + sulfur_equivalent_emissions.loc[k] / 1000
            )

            self.df_climate.loc[k, "historical_cumulative_non_co2_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "historical_cumulative_non_co2_equivalent_emissions"]
                + non_co2_equivalent_emissions.loc[k] / 1000
            )

        historical_cumulative_co2_emissions = self.df_climate["historical_cumulative_co2_emissions"]
        historical_cumulative_contrails_equivalent_emissions = self.df_climate[
            "historical_cumulative_contrails_equivalent_emissions"
        ]
        historical_cumulative_nox_short_term_o3_increase_equivalent_emissions = self.df_climate[
            "historical_cumulative_nox_short_term_o3_increase_equivalent_emissions"
        ]
        historical_cumulative_nox_long_term_o3_decrease_equivalent_emissions = self.df_climate[
            "historical_cumulative_nox_long_term_o3_decrease_equivalent_emissions"
        ]
        historical_cumulative_nox_ch4_decrease_equivalent_emissions = self.df_climate[
            "historical_cumulative_nox_ch4_decrease_equivalent_emissions"
        ]
        historical_cumulative_nox_stratospheric_water_vapor_decrease_equivalent_emissions = (
            self.df_climate[
                "historical_cumulative_nox_stratospheric_water_vapor_decrease_equivalent_emissions"
            ]
        )
        historical_cumulative_soot_equivalent_emissions = self.df_climate[
            "historical_cumulative_soot_equivalent_emissions"
        ]
        historical_cumulative_h2o_equivalent_emissions = self.df_climate[
            "historical_cumulative_h2o_equivalent_emissions"
        ]
        historical_cumulative_sulfur_equivalent_emissions = self.df_climate[
            "historical_cumulative_sulfur_equivalent_emissions"
        ]
        historical_cumulative_non_co2_equivalent_emissions = self.df_climate[
            "historical_cumulative_non_co2_equivalent_emissions"
        ]
        historical_cumulative_total_equivalent_emissions = (
            historical_cumulative_co2_emissions + historical_cumulative_non_co2_equivalent_emissions
        )
        self.df_climate[
            "cumulative_total_equivalent_emissions"
        ] = historical_cumulative_total_equivalent_emissions

        ### From 2020
        self.df_climate.loc[
            self.prospection_start_year - 1, "cumulative_non_co2_equivalent_emissions"
        ] = 0.0
        self.df_climate.loc[
            self.prospection_start_year - 1, "cumulative_total_equivalent_emissions"
        ] = 0.0
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df_climate.loc[k, "cumulative_non_co2_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "cumulative_non_co2_equivalent_emissions"]
                + non_co2_equivalent_emissions.loc[k] / 1000
            )
            self.df_climate.loc[k, "cumulative_total_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "cumulative_total_equivalent_emissions"]
                + total_equivalent_emissions.loc[k] / 1000
            )
        cumulative_non_co2_equivalent_emissions = self.df_climate[
            "cumulative_non_co2_equivalent_emissions"
        ]
        cumulative_total_equivalent_emissions = self.df_climate[
            "cumulative_total_equivalent_emissions"
        ]

        ## Share CO2/non-CO2
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            self.df_climate.loc[k, "total_co2_equivalent_emissions_ratio"] = (
                total_equivalent_emissions.loc[k] / co2_emissions.loc[k]
            )
        total_co2_equivalent_emissions_ratio = self.df_climate[
            "total_co2_equivalent_emissions_ratio"
        ]

        co2_total_erf_ratio = co2_erf / total_erf * 100
        self.df_climate.loc[:, "co2_total_erf_ratio"] = co2_total_erf_ratio

        # TEMPERATURE

        for k in range(self.climate_historic_start_year, self.end_year + 1):
            self.df_climate.loc[k, "temperature_increase_from_co2_from_aviation"] = (
                tcre_coefficient * historical_cumulative_co2_emissions.loc[k]
            )
            self.df_climate.loc[k, "temperature_increase_from_contrails_from_aviation"] = (
                tcre_coefficient * historical_cumulative_contrails_equivalent_emissions.loc[k]
            )
            self.df_climate.loc[
                k, "temperature_increase_from_nox_short_term_o3_increase_from_aviation"
            ] = (
                tcre_coefficient
                * historical_cumulative_nox_short_term_o3_increase_equivalent_emissions.loc[k]
            )
            self.df_climate.loc[
                k, "temperature_increase_from_nox_long_term_o3_decrease_from_aviation"
            ] = (
                tcre_coefficient
                * historical_cumulative_nox_long_term_o3_decrease_equivalent_emissions.loc[k]
            )
            self.df_climate.loc[k, "temperature_increase_from_nox_ch4_decrease_from_aviation"] = (
                tcre_coefficient
                * historical_cumulative_nox_ch4_decrease_equivalent_emissions.loc[k]
            )
            self.df_climate.loc[
                k, "temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation"
            ] = (
                tcre_coefficient
                * historical_cumulative_nox_stratospheric_water_vapor_decrease_equivalent_emissions.loc[
                    k
                ]
            )
            self.df_climate.loc[k, "temperature_increase_from_soot_from_aviation"] = (
                tcre_coefficient * historical_cumulative_soot_equivalent_emissions.loc[k]
            )
            self.df_climate.loc[k, "temperature_increase_from_h2o_from_aviation"] = (
                tcre_coefficient * historical_cumulative_h2o_equivalent_emissions.loc[k]
            )
            self.df_climate.loc[k, "temperature_increase_from_sulfur_from_aviation"] = (
                tcre_coefficient * historical_cumulative_sulfur_equivalent_emissions.loc[k]
            )
            self.df_climate.loc[k, "temperature_increase_from_non_co2_from_aviation"] = (
                tcre_coefficient * historical_cumulative_non_co2_equivalent_emissions.loc[k]
            )
        temperature_increase_from_co2_from_aviation = self.df_climate[
            "temperature_increase_from_co2_from_aviation"
        ]
        temperature_increase_from_contrails_from_aviation = self.df_climate[
            "temperature_increase_from_contrails_from_aviation"
        ]
        temperature_increase_from_nox_short_term_o3_increase_from_aviation = self.df_climate[
            "temperature_increase_from_nox_short_term_o3_increase_from_aviation"
        ]
        temperature_increase_from_nox_long_term_o3_decrease_from_aviation = self.df_climate[
            "temperature_increase_from_nox_long_term_o3_decrease_from_aviation"
        ]
        temperature_increase_from_nox_ch4_decrease_from_aviation = self.df_climate[
            "temperature_increase_from_nox_ch4_decrease_from_aviation"
        ]
        temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation = (
            self.df_climate[
                "temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation"
            ]
        )
        temperature_increase_from_h2o_from_aviation = self.df_climate[
            "temperature_increase_from_h2o_from_aviation"
        ]
        temperature_increase_from_soot_from_aviation = self.df_climate[
            "temperature_increase_from_soot_from_aviation"
        ]
        temperature_increase_from_sulfur_from_aviation = self.df_climate[
            "temperature_increase_from_sulfur_from_aviation"
        ]
        temperature_increase_from_non_co2_from_aviation = self.df_climate[
            "temperature_increase_from_non_co2_from_aviation"
        ]
        temperature_increase_from_aviation = (
            temperature_increase_from_co2_from_aviation
            + temperature_increase_from_non_co2_from_aviation
        )
        self.df_climate["temperature_increase_from_aviation"] = temperature_increase_from_aviation

        return (
            contrails_equivalent_emissions,
            nox_short_term_o3_increase_equivalent_emissions,
            nox_long_term_o3_decrease_equivalent_emissions,
            nox_ch4_decrease_equivalent_emissions,
            nox_stratospheric_water_vapor_decrease_equivalent_emissions,
            soot_equivalent_emissions,
            h2o_equivalent_emissions,
            sulfur_equivalent_emissions,
            non_co2_equivalent_emissions,
            cumulative_non_co2_equivalent_emissions,
            total_equivalent_emissions,
            cumulative_total_equivalent_emissions,
            total_co2_equivalent_emissions_ratio,
            co2_total_erf_ratio,
            temperature_increase_from_aviation,
            temperature_increase_from_co2_from_aviation,
            temperature_increase_from_non_co2_from_aviation,
            temperature_increase_from_contrails_from_aviation,
            temperature_increase_from_nox_short_term_o3_increase_from_aviation,
            temperature_increase_from_nox_long_term_o3_decrease_from_aviation,
            temperature_increase_from_nox_ch4_decrease_from_aviation,
            temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation,
            temperature_increase_from_h2o_from_aviation,
            temperature_increase_from_sulfur_from_aviation,
            temperature_increase_from_soot_from_aviation,
        )


class TemperatureSimpleGWPStar(AeromapsModel):
    def __init__(self, name="temperature_simple_gwpstar", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        total_erf: pd.Series = pd.Series(dtype="float64"),
        co2_erf: pd.Series = pd.Series(dtype="float64"),
        co2_emissions: pd.Series = pd.Series(dtype="float64"),
        tcre_coefficient: float = 0.0,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """Temperature calculation using equivalent emissions (with simple and smooth GWP* method) and TCRE."""

        # EQUIVALENT EMISSIONS

        # Smooth on 5 years (except for extrem values)

        self.df_climate.loc[self.climate_historic_start_year, "total_erf_smooth"] = (
            total_erf.loc[self.climate_historic_start_year]
            + total_erf.loc[self.climate_historic_start_year + 1]
            + total_erf.loc[self.climate_historic_start_year + 2]
        ) / 3
        self.df_climate.loc[self.climate_historic_start_year, "co2_erf_smooth"] = (
            co2_erf.loc[self.climate_historic_start_year]
            + co2_erf.loc[self.climate_historic_start_year + 1]
            + co2_erf.loc[self.climate_historic_start_year + 2]
        ) / 3
        self.df_climate.loc[self.climate_historic_start_year, "co2_emissions_smooth"] = (
            co2_emissions.loc[self.climate_historic_start_year]
            + co2_emissions.loc[self.climate_historic_start_year + 1]
            + co2_emissions.loc[self.climate_historic_start_year + 2]
        ) / 3

        self.df_climate.loc[self.climate_historic_start_year + 1, "total_erf_smooth"] = (
            total_erf.loc[self.climate_historic_start_year]
            + total_erf.loc[self.climate_historic_start_year + 1]
            + total_erf.loc[self.climate_historic_start_year + 2]
            + total_erf.loc[self.climate_historic_start_year + 3]
        ) / 4
        self.df_climate.loc[self.climate_historic_start_year + 1, "co2_erf_smooth"] = (
            co2_erf.loc[self.climate_historic_start_year]
            + co2_erf.loc[self.climate_historic_start_year + 1]
            + co2_erf.loc[self.climate_historic_start_year + 2]
            + co2_erf.loc[self.climate_historic_start_year + 3]
        ) / 4
        self.df_climate.loc[self.climate_historic_start_year + 1, "co2_emissions_smooth"] = (
            co2_emissions.loc[self.climate_historic_start_year]
            + co2_emissions.loc[self.climate_historic_start_year + 1]
            + co2_emissions.loc[self.climate_historic_start_year + 2]
            + co2_emissions.loc[self.climate_historic_start_year + 3]
        ) / 4

        for k in range(self.climate_historic_start_year + 2, self.end_year - 1):
            self.df_climate.loc[k, "total_erf_smooth"] = (
                total_erf.loc[k - 2]
                + total_erf.loc[k - 1]
                + total_erf.loc[k]
                + total_erf.loc[k + 1]
                + total_erf.loc[k + 2]
            ) / 5
            self.df_climate.loc[k, "co2_erf_smooth"] = (
                co2_erf.loc[k - 2]
                + co2_erf.loc[k - 1]
                + co2_erf.loc[k]
                + co2_erf.loc[k + 1]
                + co2_erf.loc[k + 2]
            ) / 5
            self.df_climate.loc[k, "co2_emissions_smooth"] = (
                co2_emissions.loc[k - 2]
                + co2_emissions.loc[k - 1]
                + co2_emissions.loc[k]
                + co2_emissions.loc[k + 1]
                + co2_emissions.loc[k + 2]
            ) / 5

        self.df_climate.loc[self.end_year - 1, "total_erf_smooth"] = (
            total_erf.loc[self.end_year - 3]
            + total_erf.loc[self.end_year - 2]
            + total_erf.loc[self.end_year - 1]
            + total_erf.loc[self.end_year]
        ) / 4
        self.df_climate.loc[self.end_year - 1, "co2_erf_smooth"] = (
            co2_erf.loc[self.end_year - 3]
            + co2_erf.loc[self.end_year - 2]
            + co2_erf.loc[self.end_year - 1]
            + co2_erf.loc[self.end_year]
        ) / 4
        self.df_climate.loc[self.end_year - 1, "co2_emissions_smooth"] = (
            co2_emissions.loc[self.end_year - 3]
            + co2_emissions.loc[self.end_year - 2]
            + co2_emissions.loc[self.end_year - 1]
            + co2_emissions.loc[self.end_year]
        ) / 4

        self.df_climate.loc[self.end_year, "total_erf_smooth"] = (
            total_erf.loc[self.end_year - 2]
            + total_erf.loc[self.end_year - 1]
            + total_erf.loc[self.end_year]
        ) / 3
        self.df_climate.loc[self.end_year, "co2_erf_smooth"] = (
            co2_erf.loc[self.end_year - 2]
            + co2_erf.loc[self.end_year - 1]
            + co2_erf.loc[self.end_year]
        ) / 3
        self.df_climate.loc[self.end_year, "co2_emissions_smooth"] = (
            co2_emissions.loc[self.end_year - 2]
            + co2_emissions.loc[self.end_year - 1]
            + co2_emissions.loc[self.end_year]
        ) / 3

        for k in range(self.climate_historic_start_year, self.end_year + 1):
            self.df_climate.loc[k, "non_co2_erf_smooth"] = (
                self.df_climate.loc[k, "total_erf_smooth"]
                - self.df_climate.loc[k, "co2_erf_smooth"]
            )

        total_erf_smooth = self.df_climate["total_erf_smooth"]
        co2_erf_smooth = self.df_climate["co2_erf_smooth"]
        co2_emissions_smooth = self.df_climate["co2_emissions_smooth"]
        non_co2_erf_smooth = self.df_climate["non_co2_erf_smooth"]

        # Equivalent emissions (Mtwe)

        ## DeltaF/Deltat
        gwpstar_variation_duration = 20
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            if k - self.climate_historic_start_year >= gwpstar_variation_duration:
                self.df_climate.loc[k, "non_co2_erf_smooth_variation"] = (
                    self.df_climate.loc[k, "non_co2_erf_smooth"]
                    - self.df_climate.loc[k - gwpstar_variation_duration, "non_co2_erf_smooth"]
                ) / gwpstar_variation_duration
            else:
                self.df_climate.loc[k, "non_co2_erf_smooth_variation"] = (
                    self.df_climate.loc[k, "non_co2_erf_smooth"] / gwpstar_variation_duration
                )
        non_co2_erf_smooth_variation = self.df_climate["non_co2_erf_smooth_variation"]

        ## Non-CO2 equivalent emissions
        climate_time_horizon = 100
        co2_agwp_h = AbsoluteGlobalWarmingPotentialCO2Function(climate_time_horizon)
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            self.df_climate.loc[k, "non_co2_equivalent_emissions"] = (
                self.df_climate.loc[k, "non_co2_erf_smooth_variation"]
                * climate_time_horizon
                / co2_agwp_h
            )
        non_co2_equivalent_emissions = self.df_climate["non_co2_equivalent_emissions"]

        ## Total equivalent emissions
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            self.df_climate.loc[k, "total_equivalent_emissions"] = (
                self.df_climate.loc[k, "non_co2_equivalent_emissions"]
                + self.df_climate.loc[k, "co2_emissions_smooth"]
            )
        total_equivalent_emissions = self.df_climate["total_equivalent_emissions"]

        ## Cumulative smooth CO2, non-CO2 and total equivalent emissions (Gtwe)

        ### From 1940
        self.df_climate.loc[
            self.climate_historic_start_year, "historical_cumulative_co2_emissions_smooth"
        ] = (co2_emissions_smooth.loc[self.climate_historic_start_year] / 1000)
        self.df_climate.loc[
            self.climate_historic_start_year, "historical_cumulative_non_co2_equivalent_emissions"
        ] = (non_co2_equivalent_emissions[self.climate_historic_start_year] / 1000)
        for k in range(self.climate_historic_start_year + 1, self.end_year + 1):
            self.df_climate.loc[k, "historical_cumulative_co2_emissions_smooth"] = (
                self.df_climate.loc[k - 1, "historical_cumulative_co2_emissions_smooth"]
                + co2_emissions_smooth.loc[k] / 1000
            )
            self.df_climate.loc[k, "historical_cumulative_non_co2_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "historical_cumulative_non_co2_equivalent_emissions"]
                + non_co2_equivalent_emissions.loc[k] / 1000
            )
        historical_cumulative_co2_emissions_smooth = self.df_climate[
            "historical_cumulative_co2_emissions_smooth"
        ]
        historical_cumulative_non_co2_equivalent_emissions = self.df_climate[
            "historical_cumulative_non_co2_equivalent_emissions"
        ]
        historical_cumulative_total_equivalent_emissions = (
            historical_cumulative_co2_emissions_smooth
            + historical_cumulative_non_co2_equivalent_emissions
        )
        self.df_climate[
            "cumulative_total_equivalent_emissions"
        ] = historical_cumulative_total_equivalent_emissions

        ### From 2020
        self.df_climate.loc[
            self.prospection_start_year - 1, "cumulative_co2_emissions_smooth"
        ] = 0.0
        self.df_climate.loc[
            self.prospection_start_year - 1, "cumulative_non_co2_equivalent_emissions"
        ] = 0.0
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df_climate.loc[k, "cumulative_co2_emissions_smooth"] = (
                self.df_climate.loc[k - 1, "cumulative_co2_emissions_smooth"]
                + co2_emissions_smooth.loc[k] / 1000
            )
            self.df_climate.loc[k, "cumulative_non_co2_equivalent_emissions"] = (
                self.df_climate.loc[k - 1, "cumulative_non_co2_equivalent_emissions"]
                + non_co2_equivalent_emissions.loc[k] / 1000
            )
        cumulative_co2_emissions_smooth = self.df_climate["cumulative_co2_emissions_smooth"]
        cumulative_non_co2_equivalent_emissions = self.df_climate[
            "cumulative_non_co2_equivalent_emissions"
        ]
        cumulative_total_equivalent_emissions = (
            cumulative_co2_emissions_smooth + cumulative_non_co2_equivalent_emissions
        )
        self.df_climate[
            "cumulative_total_equivalent_emissions"
        ] = cumulative_total_equivalent_emissions

        ## Share CO2/non-CO2
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            self.df_climate.loc[k, "total_co2_equivalent_emissions_ratio"] = (
                total_equivalent_emissions.loc[k] / co2_emissions.loc[k]
            )
        total_co2_equivalent_emissions_ratio = self.df_climate[
            "total_co2_equivalent_emissions_ratio"
        ]

        co2_total_erf_ratio = co2_erf / total_erf * 100
        self.df_climate.loc[:, "co2_total_erf_ratio"] = co2_total_erf_ratio

        # TEMPERATURE

        for k in range(self.climate_historic_start_year, self.end_year + 1):
            self.df_climate.loc[k, "temperature_increase_from_co2_from_aviation"] = (
                tcre_coefficient * historical_cumulative_co2_emissions_smooth.loc[k]
            )
            self.df_climate.loc[k, "temperature_increase_from_non_co2_from_aviation"] = (
                tcre_coefficient * historical_cumulative_non_co2_equivalent_emissions.loc[k]
            )
        temperature_increase_from_co2_from_aviation = self.df_climate[
            "temperature_increase_from_co2_from_aviation"
        ]
        temperature_increase_from_non_co2_from_aviation = self.df_climate[
            "temperature_increase_from_non_co2_from_aviation"
        ]

        temperature_increase_from_aviation = (
            temperature_increase_from_co2_from_aviation
            + temperature_increase_from_non_co2_from_aviation
        )
        self.df_climate["temperature_increase_from_aviation"] = temperature_increase_from_aviation

        return (
            total_erf_smooth,
            co2_erf_smooth,
            co2_emissions_smooth,
            cumulative_co2_emissions_smooth,
            non_co2_erf_smooth,
            non_co2_erf_smooth_variation,
            non_co2_equivalent_emissions,
            cumulative_non_co2_equivalent_emissions,
            total_equivalent_emissions,
            cumulative_total_equivalent_emissions,
            total_co2_equivalent_emissions_ratio,
            co2_total_erf_ratio,
            temperature_increase_from_aviation,
            temperature_increase_from_co2_from_aviation,
            temperature_increase_from_non_co2_from_aviation,
        )


class TemperatureFair(AeromapsModel):
    def __init__(self, name="temperature_fair", *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        # Load dataset
        rcp26_data_path = pth.join(RCP.__path__[0], "RCP26.csv")
        rcp26_data_df = pd.read_csv(rcp26_data_path)
        self.rcp26_data_df = rcp26_data_df
        rcp45_data_path = pth.join(RCP.__path__[0], "RCP45.csv")
        rcp45_data_df = pd.read_csv(rcp45_data_path)
        self.rcp45_data_df = rcp45_data_df
        rcp60_data_path = pth.join(RCP.__path__[0], "RCP60.csv")
        rcp60_data_df = pd.read_csv(rcp60_data_path)
        self.rcp60_data_df = rcp60_data_df
        rcp85_data_path = pth.join(RCP.__path__[0], "RCP85.csv")
        rcp85_data_df = pd.read_csv(rcp85_data_path)
        self.rcp85_data_df = rcp85_data_df

    def compute(
        self,
        co2_emissions: pd.Series = pd.Series(dtype="float64"),
        contrails_erf: pd.Series = pd.Series(dtype="float64"),
        nox_short_term_o3_increase_erf: pd.Series = pd.Series(dtype="float64"),
        nox_long_term_o3_decrease_erf: pd.Series = pd.Series(dtype="float64"),
        nox_ch4_decrease_erf: pd.Series = pd.Series(dtype="float64"),
        nox_stratospheric_water_vapor_decrease_erf: pd.Series = pd.Series(dtype="float64"),
        h2o_erf: pd.Series = pd.Series(dtype="float64"),
        sulfur_emissions: pd.Series = pd.Series(dtype="float64"),
        soot_emissions: pd.Series = pd.Series(dtype="float64"),
        nox_long_term_o3_decrease_gwpstar_variation_duration: float = 0.0,
        nox_long_term_o3_decrease_gwpstar_s_coefficient: float = 0.0,
        nox_ch4_decrease_gwpstar_variation_duration: float = 0.0,
        nox_ch4_decrease_gwpstar_s_coefficient: float = 0.0,
        nox_stratospheric_water_vapor_decrease_gwpstar_variation_duration: float = 0.0,
        nox_stratospheric_water_vapor_decrease_gwpstar_s_coefficient: float = 0.0,
    ) -> Tuple[
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
        pd.Series,
    ]:
        """Temperature calculation using FaIR."""

        # SPECIES QUANTITIES
        species_quantities = np.zeros((11, self.end_year - 1765 + 1))
        rcp_data_df = self.rcp45_data_df

        ## CO2

        ### World CO2
        species_quantities[0] = (
            (
                rcp_data_df["FossilCO2"][0 : self.end_year - 1765 + 1].values
                + rcp_data_df["OtherCO2"][0 : self.end_year - 1765 + 1].values
            )
            * 44
            / 12
        )  # Conversion from GtC to GtCO2

        ### Aviation CO2
        species_quantities[1] = np.zeros(len(species_quantities[0]))
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[1][k - 1765] = (
                co2_emissions.loc[k] / 1000
            )  # Conversion from MtCO2 to GtCO2

        ### Aviation NOx - Long-term O3 decrease
        species_quantities[2] = np.zeros(len(species_quantities[0]))
        nox_long_term_o3_decrease_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=nox_long_term_o3_decrease_erf,
            gwpstar_variation_duration=nox_long_term_o3_decrease_gwpstar_variation_duration,
            gwpstar_s_coefficient=nox_long_term_o3_decrease_gwpstar_s_coefficient,
        )
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[2][k - 1765] = (
                nox_long_term_o3_decrease_equivalent_emissions.loc[k] / 1000
            )  # Conversion from MtCO2-we to GtCO2-we

        ### Aviation NOx - CH4 decrease
        species_quantities[3] = np.zeros(len(species_quantities[0]))
        nox_ch4_decrease_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=nox_ch4_decrease_erf,
            gwpstar_variation_duration=nox_ch4_decrease_gwpstar_variation_duration,
            gwpstar_s_coefficient=nox_ch4_decrease_gwpstar_s_coefficient,
        )
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[3][k - 1765] = (
                nox_ch4_decrease_equivalent_emissions.loc[k] / 1000
            )  # Conversion from MtCO2-we  to GtCO2-we

        ### Aviation NOx - Stratospheric water vapor decrease
        species_quantities[4] = np.zeros(len(species_quantities[0]))
        nox_stratospheric_water_vapor_decrease_equivalent_emissions = GWPStarEquivalentEmissionsFunction(
            self,
            emissions_erf=nox_stratospheric_water_vapor_decrease_erf,
            gwpstar_variation_duration=nox_stratospheric_water_vapor_decrease_gwpstar_variation_duration,
            gwpstar_s_coefficient=nox_stratospheric_water_vapor_decrease_gwpstar_s_coefficient,
        )
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[4][k - 1765] = (
                nox_stratospheric_water_vapor_decrease_equivalent_emissions.loc[k] / 1000
            )  # Conversion from MtCO2-we  to GtCO2-we

        ## World CH4
        species_quantities[5] = rcp_data_df["CH4"][
            0 : self.end_year - 1765 + 1
        ].values  # Unit: MtCH4

        ## Aviation contrails
        species_quantities[6] = np.zeros(len(species_quantities[0]))
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[6][k - 1765] = (
                contrails_erf.loc[k] / 1000
            )  # Conversion from mW/m² to W/m²

        ## Aviation NOx - Short-term O3 increase
        species_quantities[7] = np.zeros(len(species_quantities[0]))
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[7][k - 1765] = (
                nox_short_term_o3_increase_erf.loc[k] / 1000
            )  # Conversion from mW/m² to W/m²

        ## Aviation H2O
        species_quantities[8] = np.zeros(len(species_quantities[0]))
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[8][k - 1765] = h2o_erf.loc[k] / 1000  # Conversion from mW/m² to W/m²

        ## Aviation sulfur
        species_quantities[9] = np.zeros(len(species_quantities[0]))
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[9][k - 1765] = sulfur_emissions.loc[k]  # Unit: MtSO2

        ## Aviation soot
        species_quantities[10] = np.zeros(len(species_quantities[0]))
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            species_quantities[10][k - 1765] = soot_emissions.loc[k]  # Unit: MtBC

        # TEMPERATURE ESTIMATION

        ## Total temperature (world + aviation)
        total_temperature_list = RunFair(
            self,
            species_quantities,
        )
        ## Temperature increase due to aviation species
        temperature_increase_from_co2_from_aviation_list = total_temperature_list - RunFair(
            self,
            species_quantities,
            without="Aviation CO2",
        )
        temperature_increase_from_contrails_from_aviation_list = total_temperature_list - RunFair(
            self,
            species_quantities,
            without="Aviation contrails",
        )
        temperature_increase_from_nox_short_term_o3_increase_from_aviation_list = (
            total_temperature_list
            - RunFair(
                self,
                species_quantities,
                without="Aviation NOx ST O3 increase",
            )
        )
        temperature_increase_from_nox_long_term_o3_decrease_from_aviation_list = (
            total_temperature_list
            - RunFair(
                self,
                species_quantities,
                without="Aviation NOx LT O3 decrease",
            )
        )
        temperature_increase_from_nox_ch4_decrease_from_aviation_list = (
            total_temperature_list
            - RunFair(
                self,
                species_quantities,
                without="Aviation NOx CH4 decrease",
            )
        )
        temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation_list = (
            total_temperature_list
            - RunFair(
                self,
                species_quantities,
                without="Aviation NOx H2O decrease",
            )
        )
        temperature_increase_from_h2o_from_aviation_list = total_temperature_list - RunFair(
            self,
            species_quantities,
            without="Aviation H2O",
        )
        temperature_increase_from_sulfur_from_aviation_list = total_temperature_list - RunFair(
            self,
            species_quantities,
            without="Aviation sulfur",
        )
        temperature_increase_from_soot_from_aviation_list = total_temperature_list - RunFair(
            self,
            species_quantities,
            without="Aviation soot",
        )
        # temperature_increase_from_aviation_list = total_temperature_list - RunFair(
        #     self,
        #     species_quantities,
        #     without="All aviation",
        # )

        ## List to dataframe
        for k in range(self.climate_historic_start_year, self.end_year + 1):
            self.df_climate.loc[
                k, "temperature_increase_from_co2_from_aviation"
            ] = temperature_increase_from_co2_from_aviation_list[k - 1765]
            self.df_climate.loc[
                k, "temperature_increase_from_contrails_from_aviation"
            ] = temperature_increase_from_contrails_from_aviation_list[k - 1765]
            self.df_climate.loc[
                k, "temperature_increase_from_nox_short_term_o3_increase_from_aviation"
            ] = temperature_increase_from_nox_short_term_o3_increase_from_aviation_list[k - 1765]
            self.df_climate.loc[
                k, "temperature_increase_from_nox_long_term_o3_decrease_from_aviation"
            ] = temperature_increase_from_nox_long_term_o3_decrease_from_aviation_list[k - 1765]
            self.df_climate.loc[
                k, "temperature_increase_from_nox_ch4_decrease_from_aviation"
            ] = temperature_increase_from_nox_ch4_decrease_from_aviation_list[k - 1765]
            self.df_climate.loc[
                k, "temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation"
            ] = temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation_list[
                k - 1765
            ]
            self.df_climate.loc[
                k, "temperature_increase_from_h2o_from_aviation"
            ] = temperature_increase_from_h2o_from_aviation_list[k - 1765]
            self.df_climate.loc[
                k, "temperature_increase_from_sulfur_from_aviation"
            ] = temperature_increase_from_sulfur_from_aviation_list[k - 1765]
            self.df_climate.loc[
                k, "temperature_increase_from_soot_from_aviation"
            ] = temperature_increase_from_soot_from_aviation_list[k - 1765]

        temperature_increase_from_co2_from_aviation = self.df_climate[
            "temperature_increase_from_co2_from_aviation"
        ]
        temperature_increase_from_contrails_from_aviation = self.df_climate[
            "temperature_increase_from_contrails_from_aviation"
        ]
        temperature_increase_from_nox_short_term_o3_increase_from_aviation = self.df_climate[
            "temperature_increase_from_nox_short_term_o3_increase_from_aviation"
        ]
        temperature_increase_from_nox_long_term_o3_decrease_from_aviation = self.df_climate[
            "temperature_increase_from_nox_long_term_o3_decrease_from_aviation"
        ]
        temperature_increase_from_nox_ch4_decrease_from_aviation = self.df_climate[
            "temperature_increase_from_nox_ch4_decrease_from_aviation"
        ]
        temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation = (
            self.df_climate[
                "temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation"
            ]
        )
        temperature_increase_from_h2o_from_aviation = self.df_climate[
            "temperature_increase_from_h2o_from_aviation"
        ]
        temperature_increase_from_sulfur_from_aviation = self.df_climate[
            "temperature_increase_from_sulfur_from_aviation"
        ]
        temperature_increase_from_soot_from_aviation = self.df_climate[
            "temperature_increase_from_soot_from_aviation"
        ]

        ## Temperature increase due to aviation
        temperature_increase_from_non_co2_from_aviation = (
            temperature_increase_from_contrails_from_aviation
            + temperature_increase_from_nox_short_term_o3_increase_from_aviation
            + temperature_increase_from_nox_long_term_o3_decrease_from_aviation
            + temperature_increase_from_nox_ch4_decrease_from_aviation
            + temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation
            + temperature_increase_from_h2o_from_aviation
            + temperature_increase_from_sulfur_from_aviation
            + temperature_increase_from_soot_from_aviation
        )
        temperature_increase_from_aviation = (
            temperature_increase_from_co2_from_aviation
            + temperature_increase_from_non_co2_from_aviation
        )

        self.df_climate[
            "temperature_increase_from_non_co2_from_aviation"
        ] = temperature_increase_from_non_co2_from_aviation
        self.df_climate["temperature_increase_from_aviation"] = temperature_increase_from_aviation

        return (
            temperature_increase_from_aviation,
            temperature_increase_from_co2_from_aviation,
            temperature_increase_from_non_co2_from_aviation,
            temperature_increase_from_contrails_from_aviation,
            temperature_increase_from_nox_short_term_o3_increase_from_aviation,
            temperature_increase_from_nox_long_term_o3_decrease_from_aviation,
            temperature_increase_from_nox_ch4_decrease_from_aviation,
            temperature_increase_from_nox_stratospheric_water_vapor_decrease_from_aviation,
            temperature_increase_from_h2o_from_aviation,
            temperature_increase_from_sulfur_from_aviation,
            temperature_increase_from_soot_from_aviation,
        )



from typing import Tuple

import pandas as pd

from aeromaps.models.base import AeromapsModel, AeromapsInterpolationFunction


class BiofuelProduction(AeromapsModel):
    def __init__(self, name="biofuel_production", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        biofuel_hefa_fog_share_reference_years: list = [],
        biofuel_hefa_fog_share_reference_years_values: list = [],
        biofuel_hefa_others_share_reference_years: list = [],
        biofuel_hefa_others_share_reference_years_values: list = [],
        biofuel_ft_others_share_reference_years: list = [],
        biofuel_ft_others_share_reference_years_values: list = [],
        biofuel_ft_msw_share_reference_years: list = [],
        biofuel_ft_msw_share_reference_years_values: list = [],
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        """Biomass production calculation using interpolation functions"""

        # HEFA FOG
        biofuel_hefa_fog_share = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_fog_share_reference_years,
            biofuel_hefa_fog_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_fog_share"] = biofuel_hefa_fog_share

        # HEFA OTHERS
        biofuel_hefa_others_share = AeromapsInterpolationFunction(
            self,
            biofuel_hefa_others_share_reference_years,
            biofuel_hefa_others_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_hefa_others_share"] = biofuel_hefa_others_share

        # FT OTHERS
        biofuel_ft_others_share = AeromapsInterpolationFunction(
            self,
            biofuel_ft_others_share_reference_years,
            biofuel_ft_others_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_others_share"] = biofuel_ft_others_share

        # FT MSW
        biofuel_ft_msw_share = AeromapsInterpolationFunction(
            self,
            biofuel_ft_msw_share_reference_years,
            biofuel_ft_msw_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_ft_msw_share"] = biofuel_ft_msw_share

        # ATJ
        biofuel_atj_share = (
            100
            - biofuel_hefa_fog_share
            - biofuel_hefa_others_share
            - biofuel_ft_others_share
            - biofuel_ft_msw_share
        )
        self.df.loc[:, "biofuel_atj_share"] = biofuel_atj_share

        return (
            biofuel_hefa_fog_share,
            biofuel_hefa_others_share,
            biofuel_ft_others_share,
            biofuel_ft_msw_share,
            biofuel_atj_share,
        )


class HydrogenProduction(AeromapsModel):
    def __init__(self, name="hydrogen_production", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)

    def compute(
        self,
        hydrogen_electrolysis_share_reference_years: list = [],
        hydrogen_electrolysis_share_reference_years_values: list = [],
        hydrogen_gas_ccs_share_reference_years: list = [],
        hydrogen_gas_ccs_share_reference_years_values: list = [],
        hydrogen_coal_ccs_share_reference_years: list = [],
        hydrogen_coal_ccs_share_reference_years_values: list = [],
        hydrogen_gas_share_reference_years: list = [],
        hydrogen_gas_share_reference_years_values: list = [],
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        """Hydrogen production calculation using interpolation functions"""

        # Electrolysis
        hydrogen_electrolysis_share = AeromapsInterpolationFunction(
            self,
            hydrogen_electrolysis_share_reference_years,
            hydrogen_electrolysis_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "hydrogen_electrolysis_share"] = hydrogen_electrolysis_share

        # Gas CCS
        hydrogen_gas_ccs_share = AeromapsInterpolationFunction(
            self,
            hydrogen_gas_ccs_share_reference_years,
            hydrogen_gas_ccs_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "hydrogen_gas_ccs_share"] = hydrogen_gas_ccs_share

        # Coal CCS
        hydrogen_coal_ccs_share = AeromapsInterpolationFunction(
            self,
            hydrogen_coal_ccs_share_reference_years,
            hydrogen_coal_ccs_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "hydrogen_coal_ccs_share"] = hydrogen_coal_ccs_share

        # Gas
        hydrogen_gas_share = AeromapsInterpolationFunction(
            self,
            hydrogen_gas_share_reference_years,
            hydrogen_gas_share_reference_years_values,
            model_name=self.name,
        )
        self.df.loc[:, "hydrogen_gas_share"] = hydrogen_gas_share

        # Coal
        hydrogen_coal_share = (
            100
            - hydrogen_electrolysis_share
            - hydrogen_gas_ccs_share
            - hydrogen_coal_ccs_share
            - hydrogen_gas_share
        )
        self.df.loc[:, "hydrogen_coal_share"] = hydrogen_coal_share

        return (
            hydrogen_electrolysis_share,
            hydrogen_gas_ccs_share,
            hydrogen_coal_ccs_share,
            hydrogen_gas_share,
            hydrogen_coal_share,
        )



import pandas as pd
import numpy as np
from scipy.interpolate import interp1d
import warnings
from fair import FAIR
from fair.interface import fill, initialise


class AeromapsModel(object):
    def __init__(
        self,
        name,
        parameters=None,
    ):
        self.name = name
        self.parameters = parameters
        self.float_outputs = {}
        if self.parameters is not None:
            self._initialize_df()

    def _initialize_df(self):
        self.climate_historic_start_year = self.parameters.climate_historic_start_year
        self.historic_start_year = self.parameters.historic_start_year
        self.prospection_start_year = self.parameters.prospection_start_year
        self.end_year = self.parameters.end_year
        self.df: pd.DataFrame = pd.DataFrame(
            index=range(self.historic_start_year, self.end_year + 1)
        )
        self.df_climate: pd.DataFrame = pd.DataFrame(
            index=range(self.climate_historic_start_year, self.end_year + 1)
        )
        self.years = np.linspace(self.historic_start_year, self.end_year, len(self.df.index))


def AeromapsInterpolationFunction(
    self,
    reference_years,
    reference_years_values,
    method="linear",
    positive_constraint=False,
    model_name="Not provided",
):

    # Main
    if len(reference_years) == 0:
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "interpolation_function_values"] = reference_years_values
    else:
        interpolation_function = interp1d(
            reference_years,
            reference_years_values,
            kind=method,
        )
        if reference_years[-1] == self.end_year:
            for k in range(self.prospection_start_year, reference_years[-1] + 1):
                if positive_constraint and interpolation_function(k) <= 0.0:
                    self.df.loc[k, "interpolation_function_values"] = 0.0
                else:
                    self.df.loc[k, "interpolation_function_values"] = interpolation_function(k)
        elif reference_years[-1] > self.end_year:
            warnings.warn(
                "Warning Message - "
                + "Model name: "
                + model_name
                + " - Warning on AeromapsInterpolationFunction:"
                + " The last reference year for the interpolation is higher than end_year, the interpolation function is therefore not used in its entirety.",
            )
            for k in range(self.prospection_start_year, self.end_year + 1):
                if positive_constraint and interpolation_function(k) <= 0.0:
                    self.df.loc[k, "interpolation_function_values"] = 0.0
                else:
                    self.df.loc[k, "interpolation_function_values"] = interpolation_function(k)
        else:
            warnings.warn(
                "Warning Message - "
                + "Model name: "
                + model_name
                + " - Warning on AeromapsInterpolationFunction:"
                + " The last reference year for the interpolation is lower than end_year, the value associated to the last reference year is therefore used as a constant for the upper years.",
            )
            for k in range(self.prospection_start_year, reference_years[-1] + 1):
                if positive_constraint and interpolation_function(k) <= 0.0:
                    self.df.loc[k, "interpolation_function_values"] = 0.0
                else:
                    self.df.loc[k, "interpolation_function_values"] = interpolation_function(k)
            for k in range(reference_years[-1] + 1, self.end_year + 1):
                self.df.loc[k, "interpolation_function_values"] = self.df.loc[
                    k - 1, "interpolation_function_values"
                ]

    interpolation_function_values = self.df.loc[:, "interpolation_function_values"]

    # Delete intermediate df column
    self.df.pop("interpolation_function_values")

    return interpolation_function_values


def AeromapsLevelingFunction(
    self, reference_periods, reference_periods_values, model_name="Not provided"
):

    # Main
    if len(reference_periods) == 0:
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df.loc[k, "leveling_function_values"] = reference_periods_values
    else:
        if reference_periods[-1] == self.end_year:
            for i in range(0, len(reference_periods) - 1):
                for k in range(reference_periods[i], reference_periods[i + 1] + 1):
                    self.df.loc[k, "leveling_function_values"] = reference_periods_values[i]
        elif reference_periods[-1] > self.end_year:
            warnings.warn(
                "Warning Message - "
                + "Model name: "
                + model_name
                + " - Warning on AeromapsLevelingFunction:"
                + " The last reference year for the leveling is higher than end_year, the leveling function is therefore not used in its entirety.",
            )
            for i in range(0, len(reference_periods) - 1):
                for k in range(reference_periods[i], reference_periods[i + 1] + 1):
                    if k <= self.end_year:
                        self.df.loc[k, "leveling_function_values"] = reference_periods_values[i]
                    else:
                        pass
        else:
            warnings.warn(
                "Warning Message - "
                + "Model name: "
                + model_name
                + " - Warning on AeromapsLevelingFunction:"
                + " The last reference year for the leveling is lower than end_year, the value associated to the last reference period is therefore used as a constant for the upper period.",
            )
            for i in range(0, len(reference_periods) - 1):
                for k in range(reference_periods[i], reference_periods[i + 1] + 1):
                    self.df.loc[k, "leveling_function_values"] = reference_periods_values[i]
            for k in range(reference_periods[-1] + 1, self.end_year + 1):
                self.df.loc[k, "leveling_function_values"] = self.df.loc[
                    k - 1, "leveling_function_values"
                ]

    leveling_function_values = self.df.loc[:, "leveling_function_values"]

    # Delete intermediate df column
    self.df.pop("leveling_function_values")

    return leveling_function_values


def AbsoluteGlobalWarmingPotentialCO2Function(climate_time_horizon):

    # Reference: IPCC AR5 - https://www.ipcc.ch/site/assets/uploads/2018/07/WGI_AR5.Chap_.8_SM.pdf

    # Parameter: climate time horizon
    h = climate_time_horizon

    co2_molar_mass = 44.01 * 1e-3  # [kg/mol]
    air_molar_mass = 28.97e-3  # [kg/mol]
    atmosphere_total_mass = 5.1352e18  # [kg]

    radiative_efficiency = 1.37e-2 * 1e9  # radiative efficiency [mW/m^2]

    # RF per unit mass increase in atmospheric abundance of CO2 [W/m^2/kg]
    A_CO2 = radiative_efficiency * air_molar_mass / (co2_molar_mass * atmosphere_total_mass) * 1e-3

    # Coefficients for the model
    a = [0.2173, 0.2240, 0.2824, 0.2763]
    tau = [0, 394.4, 36.54, 4.304]  # CO2 lifetime [yrs]

    co2_agwp_h = A_CO2 * a[0] * h
    for i in [1, 2, 3]:
        co2_agwp_h += A_CO2 * a[i] * tau[i] * (1 - np.exp(-h / tau[i]))

    # From W/m^2/kg.yr to mW/m^2/Mt.yr
    co2_agwp_h = co2_agwp_h * 1e3 * 1e9

    return co2_agwp_h


def GWPStarEquivalentEmissionsFunction(
    self, emissions_erf, gwpstar_variation_duration, gwpstar_s_coefficient
):

    # Reference: Smith et al. (2021), https://doi.org/10.1038/s41612-021-00169-8
    # Global
    climate_time_horizon = 100
    co2_agwp_h = AbsoluteGlobalWarmingPotentialCO2Function(climate_time_horizon)

    # g coefficient for GWP*
    if gwpstar_s_coefficient == 0:
        g_coefficient = 1
    else:
        g_coefficient = (
            1 - np.exp(-gwpstar_s_coefficient / (1 - gwpstar_s_coefficient))
        ) / gwpstar_s_coefficient

    # Main
    for k in range(self.climate_historic_start_year, self.end_year + 1):
        if k - self.climate_historic_start_year >= gwpstar_variation_duration:
            self.df_climate.loc[k, "emissions_erf_variation"] = (
                emissions_erf.loc[k] - emissions_erf.loc[k - gwpstar_variation_duration]
            ) / gwpstar_variation_duration
        else:
            self.df_climate.loc[k, "emissions_erf_variation"] = (
                emissions_erf.loc[k] / gwpstar_variation_duration
            )

    for k in range(self.climate_historic_start_year, self.end_year + 1):
        self.df_climate.loc[k, "emissions_equivalent_emissions"] = (
            g_coefficient
            * (1 - gwpstar_s_coefficient)
            * climate_time_horizon
            / co2_agwp_h
            * self.df_climate.loc[k, "emissions_erf_variation"]
        ) + g_coefficient * gwpstar_s_coefficient / co2_agwp_h * emissions_erf.loc[k]
    emissions_equivalent_emissions = self.df_climate.loc[:, "emissions_equivalent_emissions"]

    # Delete intermediate df column
    self.df_climate.pop("emissions_erf_variation")
    self.df_climate.pop("emissions_equivalent_emissions")

    return emissions_equivalent_emissions


def RunFair(self, species_quantities, without="None"):

    # Creation of FaIR instance
    f = FAIR()

    # Definition of time horizon, scenarios, configs
    start_time = 1765
    end_time = self.end_year
    f.define_time(start_time, end_time, 1)
    f.define_scenarios(["central"])
    f.define_configs(["central"])
    # f.define_configs(["high", "central", "low"])

    # Definition of species and properties
    species = [
        "CO2",  # Includes world emissions, aviation emissions, and equivalent emissions for NOx effects (except ST O3)
        "World CH4",
        "Aviation contrails",
        "Aviation NOx ST O3 increase",
        "Aviation H2O",
        "Aviation sulfur",
        "Aviation soot",
        "Aviation aerosols",
    ]
    properties = {
        "CO2": {
            "type": "co2",
            "input_mode": "emissions",
            "greenhouse_gas": True,
            "aerosol_chemistry_from_emissions": False,
            "aerosol_chemistry_from_concentration": False,
        },
        "World CH4": {
            "type": "ch4",
            "input_mode": "emissions",
            "greenhouse_gas": True,
            "aerosol_chemistry_from_emissions": False,
            "aerosol_chemistry_from_concentration": True,  # we treat methane as a reactive gas
        },
        "Aviation contrails": {
            "type": "contrails",
            "input_mode": "forcing",
            "greenhouse_gas": False,
            "aerosol_chemistry_from_emissions": False,
            "aerosol_chemistry_from_concentration": False,
        },
        "Aviation NOx ST O3 increase": {
            "type": "ozone",
            "input_mode": "forcing",
            "greenhouse_gas": False,
            "aerosol_chemistry_from_emissions": False,
            "aerosol_chemistry_from_concentration": False,
        },
        "Aviation H2O": {
            "type": "h2o stratospheric",
            "input_mode": "forcing",
            "greenhouse_gas": False,
            "aerosol_chemistry_from_emissions": False,
            "aerosol_chemistry_from_concentration": False,
        },
        "Aviation sulfur": {
            "type": "sulfur",
            "input_mode": "emissions",
            "greenhouse_gas": False,
            "aerosol_chemistry_from_emissions": True,
            "aerosol_chemistry_from_concentration": False,
        },
        "Aviation soot": {
            "type": "black carbon",
            "input_mode": "emissions",
            "greenhouse_gas": False,
            "aerosol_chemistry_from_emissions": True,
            "aerosol_chemistry_from_concentration": False,
        },
        # Dedicated specie for aerosols
        "Aviation aerosols": {
            "type": "ari",
            "input_mode": "calculated",
            "greenhouse_gas": False,
            "aerosol_chemistry_from_emissions": False,
            "aerosol_chemistry_from_concentration": False,
        },
    }
    f.define_species(species, properties)

    # Definition of run options
    f.ghg_method = "leach2021"
    f.aci_method = "myhre1998"

    # Creation of input and output data
    f.allocate()

    # Filling species quantities
    if without == "All aviation":
        fill(
            f.emissions,
            species_quantities[0][1 : self.end_year - 1765 + 1],
            specie="CO2",
            config=f.configs[0],
            scenario=f.scenarios[0],
        )
        fill(
            f.emissions,
            species_quantities[5][1 : self.end_year - 1765 + 1],
            specie="World CH4",
            config=f.configs[0],
            scenario=f.scenarios[0],
        )
        fill(
            f.forcing, 0, specie="Aviation contrails", config=f.configs[0], scenario=f.scenarios[0]
        )
        fill(
            f.forcing,
            0,
            specie="Aviation NOx ST O3 increase",
            config=f.configs[0],
            scenario=f.scenarios[0],
        )
        fill(f.forcing, 0, specie="Aviation H2O", config=f.configs[0], scenario=f.scenarios[0])
        fill(f.emissions, 0, specie="Aviation sulfur", config=f.configs[0], scenario=f.scenarios[0])
        fill(f.emissions, 0, specie="Aviation soot", config=f.configs[0], scenario=f.scenarios[0])
    else:
        if without == "Aviation CO2":
            total_CO2 = (
                species_quantities[0][1 : self.end_year - 1765 + 1]
                + species_quantities[2][1 : self.end_year - 1765 + 1]
                + species_quantities[3][1 : self.end_year - 1765 + 1]
                + species_quantities[4][1 : self.end_year - 1765 + 1]
            )
        elif without == "Aviation NOx LT O3 decrease":
            total_CO2 = (
                species_quantities[0][1 : self.end_year - 1765 + 1]
                + species_quantities[1][1 : self.end_year - 1765 + 1]
                + species_quantities[3][1 : self.end_year - 1765 + 1]
                + species_quantities[4][1 : self.end_year - 1765 + 1]
            )
        elif without == "Aviation NOx CH4 decrease":
            total_CO2 = (
                species_quantities[0][1 : self.end_year - 1765 + 1]
                + species_quantities[1][1 : self.end_year - 1765 + 1]
                + species_quantities[2][1 : self.end_year - 1765 + 1]
                + species_quantities[4][1 : self.end_year - 1765 + 1]
            )
        elif without == "Aviation NOx H2O decrease":
            total_CO2 = (
                species_quantities[0][1 : self.end_year - 1765 + 1]
                + species_quantities[1][1 : self.end_year - 1765 + 1]
                + species_quantities[2][1 : self.end_year - 1765 + 1]
                + species_quantities[3][1 : self.end_year - 1765 + 1]
            )
        else:
            total_CO2 = (
                species_quantities[0][1 : self.end_year - 1765 + 1]
                + species_quantities[1][1 : self.end_year - 1765 + 1]
                + species_quantities[2][1 : self.end_year - 1765 + 1]
                + species_quantities[3][1 : self.end_year - 1765 + 1]
                + species_quantities[4][1 : self.end_year - 1765 + 1]
            )
        fill(f.emissions, total_CO2, specie="CO2", config=f.configs[0], scenario=f.scenarios[0])
        fill(
            f.emissions,
            species_quantities[5][1 : self.end_year - 1765 + 1],
            specie="World CH4",
            config=f.configs[0],
            scenario=f.scenarios[0],
        )
        if without != "Aviation contrails":
            fill(
                f.forcing,
                species_quantities[6],
                specie="Aviation contrails",
                config=f.configs[0],
                scenario=f.scenarios[0],
            )
        else:
            fill(
                f.forcing,
                0,
                specie="Aviation contrails",
                config=f.configs[0],
                scenario=f.scenarios[0],
            )
        if without != "Aviation NOx ST O3 increase":
            fill(
                f.forcing,
                species_quantities[7],
                specie="Aviation NOx ST O3 increase",
                config=f.configs[0],
                scenario=f.scenarios[0],
            )
        else:
            fill(
                f.forcing,
                0,
                specie="Aviation NOx ST O3 increase",
                config=f.configs[0],
                scenario=f.scenarios[0],
            )
        if without != "Aviation H2O":
            fill(
                f.forcing,
                species_quantities[8],
                specie="Aviation H2O",
                config=f.configs[0],
                scenario=f.scenarios[0],
            )
        else:
            fill(f.forcing, 0, specie="Aviation H2O", config=f.configs[0], scenario=f.scenarios[0])
        if without != "Aviation sulfur":
            fill(
                f.emissions,
                species_quantities[9][1 : self.end_year - 1765 + 1],
                specie="Aviation sulfur",
                config=f.configs[0],
                scenario=f.scenarios[0],
            )
        else:
            fill(
                f.emissions,
                0,
                specie="Aviation sulfur",
                config=f.configs[0],
                scenario=f.scenarios[0],
            )
        if without != "Aviation soot":
            fill(
                f.emissions,
                species_quantities[10][1 : self.end_year - 1765 + 1],
                specie="Aviation soot",
                config=f.configs[0],
                scenario=f.scenarios[0],
            )
        else:
            fill(
                f.emissions, 0, specie="Aviation soot", config=f.configs[0], scenario=f.scenarios[0]
            )

    initialise(f.forcing, 0)
    initialise(f.temperature, 0)
    initialise(f.cumulative_emissions, 0)
    initialise(f.airborne_emissions, 0)

    # Filling climate configs
    # fill(f.climate_configs["ocean_heat_transfer"], [1.1, 1.6, 0.9], config="central")
    # fill(f.climate_configs["ocean_heat_capacity"], [8, 14, 100], config="central")
    # fill(f.climate_configs["deep_ocean_efficacy"], 1.1, config="central")
    # Corresponds to a "low" configuration on FaIR
    fill(f.climate_configs["ocean_heat_transfer"], [1.7, 2.0, 1.1], config="central")
    fill(f.climate_configs["ocean_heat_capacity"], [6, 11, 75], config="central")
    fill(f.climate_configs["deep_ocean_efficacy"], 0.8, config="central")

    # Filling species configs
    for specie in species:
        if specie == "CO2":
            fill(
                f.species_configs["partition_fraction"],
                [0.2173, 0.2240, 0.2824, 0.2763],
                specie="CO2",
            )
            fill(
                f.species_configs["unperturbed_lifetime"], [1e9, 394.4, 36.54, 4.304], specie="CO2"
            )
            fill(f.species_configs["baseline_concentration"], 278.3, specie="CO2")
            fill(f.species_configs["forcing_reference_concentration"], 278.3, specie="CO2")
            fill(f.species_configs["molecular_weight"], 44.009, specie="CO2")
            fill(
                f.species_configs["greenhouse_gas_radiative_efficiency"],
                1.3344985680386619e-05,
                specie="CO2",
            )
            f.calculate_iirf0()
            f.calculate_g()
            f.calculate_concentration_per_emission()
            fill(f.species_configs["iirf_0"], 29, specie="CO2")
            fill(f.species_configs["iirf_airborne"], [0.000819 * 2], specie="CO2")
            fill(f.species_configs["iirf_uptake"], [0.00846 * 2], specie="CO2")
            fill(f.species_configs["iirf_temperature"], [8], specie="CO2")
            fill(f.species_configs["aci_scale"], -2.09841432)

        if specie == "World CH4":
            fill(f.species_configs["partition_fraction"], [1, 0, 0, 0], specie=specie)
            fill(f.species_configs["unperturbed_lifetime"], 8.25, specie=specie)
            fill(f.species_configs["baseline_concentration"], 729, specie=specie)  # ppb
            fill(f.species_configs["forcing_reference_concentration"], 729, specie=specie)
            fill(f.species_configs["molecular_weight"], 16.043, specie=specie)
            fill(
                f.species_configs["greenhouse_gas_radiative_efficiency"],
                0.00038864402860869495,
                specie=specie,
            )
            f.calculate_iirf0()
            f.calculate_g()
            f.calculate_concentration_per_emission()
            fill(f.species_configs["iirf_airborne"], 0.00032, specie=specie)
            fill(f.species_configs["iirf_uptake"], 0, specie=specie)
            fill(f.species_configs["iirf_temperature"], -0.3, specie=specie)
            fill(
                f.species_configs["erfari_radiative_efficiency"],
                -0.002653 / 1023.2219696044921,
                specie=specie,
            )  # W m-2 ppb-1
            fill(f.species_configs["aci_scale"], -2.09841432)

        if specie == "Aviation sulfur":
            erf_aci_sulfur = 0.0
            fill(
                f.species_configs["erfari_radiative_efficiency"],
                -0.0199 + erf_aci_sulfur,
                specie=specie,
            )  # W m-2 MtSO2-1 yr
            fill(f.species_configs["aci_shape"], 0.0, specie=specie)

        if specie == "Aviation soot":
            erf_aci_BC = 0.0
            fill(
                f.species_configs["erfari_radiative_efficiency"], 0.1007 + erf_aci_BC, specie=specie
            )  # W m-2 MtC-1 yr
            fill(f.species_configs["aci_shape"], 0.0, specie=specie)

    # Run
    f.run()

    return f.temperature.loc[dict(layer=0, config=f.configs[0])].data



from typing import Tuple

import pandas as pd

from aeromaps.models.base import AeromapsModel, AeromapsInterpolationFunction


class DropinFuelDistribution(AeromapsModel):
    def __init__(self, name="dropin_fuel_distribution", *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    def compute(
        self,
        biofuel_share_reference_years: list = [],
        biofuel_share_reference_years_values: list = [],
        electrofuel_share_reference_years: list = [],
        electrofuel_share_reference_years_values: list = [],
    ) -> Tuple[pd.Series, pd.Series, pd.Series]:
        """Fuel distribution calculation using interpolation functions"""

        # Biofuel
        biofuel_share_prospective = AeromapsInterpolationFunction(
            self,
            biofuel_share_reference_years,
            biofuel_share_reference_years_values,
            method="quadratic",
            positive_constraint=True,
            model_name=self.name,
        )
        self.df.loc[:, "biofuel_share"] = biofuel_share_prospective
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "biofuel_share"] = self.df.loc[
                self.prospection_start_year, "biofuel_share"
            ]
        biofuel_share = self.df["biofuel_share"]

        # Electrofuel
        electrofuel_share_prospective = AeromapsInterpolationFunction(
            self,
            electrofuel_share_reference_years,
            electrofuel_share_reference_years_values,
            method="quadratic",
            positive_constraint=True,
            model_name=self.name,
        )
        self.df.loc[:, "electrofuel_share"] = electrofuel_share_prospective
        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df.loc[k, "electrofuel_share"] = self.df.loc[
                self.prospection_start_year, "electrofuel_share"
            ]
        electrofuel_share = self.df["electrofuel_share"]

        # Kerosene
        kerosene_share = 100 - biofuel_share - electrofuel_share
        self.df.loc[:, "kerosene_share"] = kerosene_share

        return biofuel_share, electrofuel_share, kerosene_share



from typing import Tuple
import pandas as pd
from pandas import read_csv
import os.path as pth

from aeromaps.models.base import AeromapsModel
from aeromaps.resources import climate_data


class TotalAircraftDistance(AeromapsModel):
    def __init__(self, name="total_aircraft_distance", *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        # Load dataset
        historical_dataset_path = pth.join(
            climate_data.__path__[0], "temperature_historical_dataset.csv"
        )
        historical_dataset_df = read_csv(historical_dataset_path, delimiter=";", header=None)
        self.historical_dataset = historical_dataset_df.values

    def compute(
        self,
        rtk: pd.Series = pd.Series(dtype="float64"),
        ask: pd.Series = pd.Series(dtype="float64"),
        ask_dropin_fuel: pd.Series = pd.Series(dtype="float64"),
        ask_hydrogen: pd.Series = pd.Series(dtype="float64"),
        ask_electric: pd.Series = pd.Series(dtype="float64"),
        total_aircraft_distance_init: pd.Series = pd.Series(dtype="float64"),
    ) -> Tuple[pd.Series, pd.Series, pd.Series, pd.Series]:
        """Total aircraft distance calculation."""

        historical_distance_for_temperature = self.historical_dataset[:, 6]

        for k in range(self.climate_historic_start_year, self.historic_start_year):
            self.df_climate.loc[k, "total_aircraft_distance"] = historical_distance_for_temperature[
                k - self.climate_historic_start_year
            ]

        for k in range(self.historic_start_year, self.prospection_start_year):
            self.df_climate.loc[k, "total_aircraft_distance"] = total_aircraft_distance_init.loc[k]

        # Assumption: 1 RTK = 10 ASK
        for k in range(self.prospection_start_year, self.end_year + 1):
            self.df_climate.loc[k, "total_aircraft_distance"] = (
                (ask.loc[k] + 10 * rtk.loc[k])
                / (
                    ask.loc[self.prospection_start_year - 1]
                    + 10 * rtk.loc[self.prospection_start_year - 1]
                )
                * self.df_climate.loc[self.prospection_start_year - 1, "total_aircraft_distance"]
            )

        total_aircraft_distance = self.df_climate["total_aircraft_distance"]

        # Assumption: distribution proportional to ASK
        total_aircraft_distance_dropin_fuel = total_aircraft_distance * ask_dropin_fuel / ask
        total_aircraft_distance_hydrogen = total_aircraft_distance * ask_hydrogen / ask
        total_aircraft_distance_electric = total_aircraft_distance * ask_electric / ask
        self.df.loc[:, "total_aircraft_distance_dropin_fuel"] = total_aircraft_distance_dropin_fuel
        self.df.loc[:, "total_aircraft_distance_hydrogen"] = total_aircraft_distance_hydrogen
        self.df.loc[:, "total_aircraft_distance_electric"] = total_aircraft_distance_electric

        return (
            total_aircraft_distance,
            total_aircraft_distance_dropin_fuel,
            total_aircraft_distance_hydrogen,
            total_aircraft_distance_electric,
        )
